# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typing
import warnings
from typing import Optional, Union

import mlrun
import mlrun.common.constants as mlrun_constants
import mlrun.common.schemas.function
import mlrun.common.schemas.workflow
import mlrun_pipelines.models
from mlrun.utils import hub_prefix

from .pipelines import enrich_function_object, pipeline_context


def _get_engine_and_function(function, project=None):
    is_function_object = not isinstance(function, str)
    project = project or pipeline_context.project
    if not is_function_object:
        if function.startswith(hub_prefix):
            function = mlrun.import_function(function)
            if project:
                function = enrich_function_object(project, function)
        else:
            if not project:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    "function name (str) can only be used in a project context, you must create, "
                    "load or get a project first or provide function object instead of its name"
                )
            # we don't want to use a copy of the function object, we want to use the actual object
            # so changes on it will be reflected in the project and will persist for future use of the function
            function = project.get_function(
                function, sync=False, enrich=True, copy_function=False
            )
    elif project:
        # if a user provide the function object we enrich in-place so build, deploy, etc.
        # will update the original function object status/image, and not the copy (may fail fn.run())
        function = enrich_function_object(project, function, copy_function=False)

    if not pipeline_context.workflow:
        return mlrun.common.schemas.workflow.EngineType.LOCAL, function

    return pipeline_context.workflow.engine, function


def run_function(
    function: Union[str, mlrun.runtimes.BaseRuntime],
    handler: Optional[Union[str, typing.Callable]] = None,
    name: str = "",
    params: Optional[dict] = None,
    hyperparams: Optional[dict] = None,
    hyper_param_options: mlrun.model.HyperParamOptions = None,
    inputs: Optional[dict] = None,
    outputs: Optional[list[str]] = None,
    workdir: str = "",
    labels: Optional[dict] = None,
    base_task: mlrun.model.RunTemplate = None,
    watch: bool = True,
    local: Optional[bool] = None,
    verbose: Optional[bool] = None,
    selector: Optional[str] = None,
    project_object=None,
    auto_build: Optional[bool] = None,
    schedule: Union[str, mlrun.common.schemas.ScheduleCronTrigger] = None,
    artifact_path: Optional[str] = None,
    notifications: Optional[list[mlrun.model.Notification]] = None,
    returns: Optional[list[Union[str, dict[str, str]]]] = None,
    builder_env: Optional[list] = None,
    reset_on_run: Optional[bool] = None,
    output_path: Optional[str] = None,
    retry: Optional[Union[mlrun.model.Retry, dict]] = None,
) -> Union[mlrun.model.RunObject, mlrun_pipelines.models.PipelineNodeWrapper]:
    """Run a local or remote task as part of a local/kubeflow pipeline

    run_function() allow you to execute a function locally, on a remote cluster, or as part of an automated workflow
    function can be specified as an object or by name (str), when the function is specified by name it is looked up
    in the current project eliminating the need to redefine/edit functions.

    when functions run as part of a workflow/pipeline (project.run()) some attributes can be set at the run level,
    e.g. local=True will run all the functions locally, setting artifact_path will direct all outputs to the same path.
    project runs provide additional notifications/reporting and exception handling.
    inside a Kubeflow pipeline (KFP) run_function() generates KFP node (see PipelineNodeWrapper) which forms a DAG
    some behavior may differ between regular runs and deferred KFP runs.

    example (use with function object)::

        LABELS = "is_error"
        MODEL_CLASS = "sklearn.ensemble.RandomForestClassifier"
        DATA_PATH = "s3://bigdata/data.parquet"
        function = mlrun.import_function("hub://auto-trainer")
        run1 = run_function(
            function,
            params={"label_columns": LABELS, "model_class": MODEL_CLASS},
            inputs={"dataset": DATA_PATH},
        )

    example (use with project)::

        # create a project with two functions (local and from hub)
        project = mlrun.new_project(project_name, "./proj)
        project.set_function("mycode.py", "myfunc", image="mlrun/mlrun")
        project.set_function("hub://auto-trainer", "train")

        # run functions (refer to them by name)
        run1 = run_function("myfunc", params={"x": 7})
        run2 = run_function("train", params={"label_columns": LABELS, "model_class": MODEL_CLASS},
                                     inputs={"dataset": run1.outputs["data"]})

    example (use in pipeline)::

        @dsl.pipeline(name="test pipeline", description="test")
        def my_pipe(url=""):
            run1 = run_function("loaddata", params={"url": url}, outputs=["data"])
            run2 = run_function(
                "train",
                params={"label_columns": LABELS, "model_class": MODEL_CLASS},
                inputs={"dataset": run1.outputs["data"]},
            )


        project.run(workflow_handler=my_pipe, arguments={"param1": 7})

    :param function:        name of the function (in the project) or function object
    :param handler:         name of the function handler
    :param name:            execution name
    :param params:          input parameters (dict)
    :param hyperparams:     hyper parameters
    :param selector:        selection criteria for hyper params e.g. "max.accuracy"
    :param hyper_param_options:  hyper param options (selector, early stop, strategy, ..)
                            see: :py:class:`~mlrun.model.HyperParamOptions`
    :param inputs:          Input objects to pass to the handler. Type hints can be given so the input will be parsed
                            during runtime from `mlrun.DataItem` to the given type hint. The type hint can be given
                            in the key field of the dictionary after a colon, e.g: "<key> : <type_hint>".
    :param outputs:         list of outputs which can pass in the workflow
    :param workdir:         working directory of the executed job and the default path for artifact inputs
    :param labels:          labels to tag the job/run with ({key:val, ..})
    :param base_task:       task object to use as base
    :param watch:           watch/follow run log, True by default
    :param local:           run the function locally vs on the runtime/cluster
    :param verbose:         add verbose prints/logs
    :param project_object:  override the project object to use, will default to the project set in the runtime context.
    :param auto_build:      when set to True and the function require build it will be built on the first
                            function run, use only if you do not plan on changing the build config between runs
    :param schedule:        ScheduleCronTrigger class instance or a standard crontab expression string
                            (which will be converted to the class using its `from_crontab` constructor),
                            see this link for help:
                            https://apscheduler.readthedocs.io/en/3.x/modules/triggers/cron.html#module-apscheduler.triggers.cron
    :param artifact_path:   (deprecated) path to store artifacts, when running in a workflow this will be set
                            automatically
    :param notifications:   list of notifications to push when the run is completed
    :param returns:         List of log hints - configurations for how to log the returning values from the handler's
                            run (as artifacts or results). The list's length must be equal to the amount of returning
                            objects. A log hint may be given as:

                            * A string of the key to use to log the returning value as result or as an artifact. To
                              specify The artifact type, it is possible to pass a string in the following structure:
                              "<key> : <type>". Available artifact types can be seen in `mlrun.ArtifactType`. If no
                              artifact type is specified, the object's default artifact type will be used.
                            * A dictionary of configurations to use when logging. Further info per object type and
                              artifact type can be given there. The artifact key must appear in the dictionary as
                              "key": "the_key".
    :param builder_env:     env vars dict for source archive config/credentials e.g. builder_env={"GIT_TOKEN": token}
    :param reset_on_run:    When True, function python modules would reload prior to code execution.
                            This ensures latest code changes are executed. This argument must be used in
                            conjunction with the local=True argument.
    :param output_path:     path to store artifacts, when running in a workflow this will be set automatically
    :param retry:           Retry configuration for the run, can be a dict or an instance of
                            :py:class:`~mlrun.model.Retry`.
                            The `count` field in the `Retry` object specifies the number of retry attempts.
                            If `count=0`, the run will not be retried.
                            The `backoff` field specifies the retry backoff strategy between retry attempts.
                            If not provided, the default backoff delay is 30 seconds.
    :return: MLRun RunObject or PipelineNodeWrapper
    """
    if artifact_path:
        warnings.warn(
            "'artifact_path' parameter is deprecated in 1.10.0 and will be removed in 1.12.0, "
            "use 'output_path' instead.",
            # TODO: Remove this in 1.12.0
            FutureWarning,
        )
    output_path = output_path or artifact_path
    engine, function = _get_engine_and_function(function, project_object)
    task = mlrun.new_task(
        handler=handler,
        params=params,
        hyper_params=hyperparams,
        hyper_param_options=hyper_param_options,
        inputs=inputs,
        returns=returns,
        base=base_task,
        selector=selector,
        retry=retry,
    )
    task.spec.verbose = task.spec.verbose or verbose

    if engine == mlrun.common.schemas.workflow.EngineType.KFP:
        if schedule:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "Scheduling jobs is not supported when running a workflow with the kfp engine."
            )
        if retry:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "Retrying jobs is not supported when running a workflow with the kfp engine. "
                "Use KFP set_retry instead."
            )
        return function.as_step(
            name=name, runspec=task, workdir=workdir, outputs=outputs, labels=labels
        )
    else:
        project = project_object or pipeline_context.project
        local = pipeline_context.is_run_local(local)
        task.metadata.labels = task.metadata.labels or labels or {}
        if pipeline_context.workflow_id:
            task.metadata.labels[mlrun_constants.MLRunInternalLabels.workflow] = (
                pipeline_context.workflow_id
            )
        if function.kind == "local":
            command, function = mlrun.run.load_func_code(function)
            function.spec.command = command
        if local and project and function.spec.build.source:
            workdir = workdir or project.spec.get_code_path()

        # remove this filter once the artifact_path parameter is deprecated in 1.12.0
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=FutureWarning)
            run_result = function.run(
                name=name,
                runspec=task,
                workdir=workdir,
                verbose=verbose,
                watch=watch,
                local=local,
                output_path=output_path
                # workflow output_path has precedence over the project artifact_path equivalent to passing
                # output_path to function.run() has precedence over the project.artifact_path and the default one
                or pipeline_context.workflow_artifact_path
                or (project.artifact_path if project else None),
                auto_build=auto_build,
                schedule=schedule,
                notifications=notifications,
                builder_env=builder_env,
                reset_on_run=reset_on_run,
            )
        if run_result:
            run_result._notified = False
            pipeline_context.runs_map[run_result.uid()] = run_result
            run_result.after = (
                lambda x: run_result
            )  # emulate KFP op, .after() will be ignored
        return run_result


class BuildStatus:
    """returned status from build operation"""

    def __init__(self, ready, outputs=None, function=None):
        self.ready = ready
        self.outputs = outputs or {}
        self.function = function

    def after(self, step):
        """nil function, for KFP compatibility"""
        return self

    def __repr__(self):
        return f"BuildStatus(ready={self.ready}, outputs={self.outputs})"


def build_function(
    function: Union[str, mlrun.runtimes.BaseRuntime],
    with_mlrun: Optional[bool] = None,
    skip_deployed: bool = False,
    image=None,
    base_image=None,
    commands: Optional[list] = None,
    secret_name=None,
    requirements: Optional[Union[str, list[str]]] = None,
    requirements_file: Optional[str] = None,
    mlrun_version_specifier=None,
    builder_env: Optional[dict] = None,
    project_object=None,
    overwrite_build_params: bool = True,
    extra_args: Optional[str] = None,
    force_build: bool = False,
) -> Union[BuildStatus, mlrun_pipelines.models.PipelineNodeWrapper]:
    """deploy ML function, build container with its dependencies

    :param function:        Name of the function (in the project) or function object
    :param with_mlrun:      Add the current mlrun package to the container build
    :param skip_deployed:   Skip the build if we already have an image for the function
    :param image:           Target image name/path
    :param base_image:      Base image name/path (commands and source code will be added to it)
    :param commands:        List of docker build (RUN) commands e.g. ['pip install pandas']
    :param secret_name:     K8s secret for accessing the docker registry
    :param requirements:    List of python packages, defaults to None
    :param requirements_file:    pip requirements file path, defaults to None
    :param mlrun_version_specifier:  which mlrun package version to include (if not current)
    :param builder_env:     Kaniko builder pod env vars dict (for config/credentials)
        e.g. builder_env={"GIT_TOKEN": token}, does not work yet in KFP
    :param project_object:  Override the project object to use, will default to the project set in the runtime context.
    :param overwrite_build_params:  Overwrite existing build configuration (currently applies to
        requirements and commands)
        * False: The new params are merged with the existing
        * True: The existing params are replaced by the new ones
    :param extra_args:  A string containing additional builder arguments in the format of command-line options,
        e.g. extra_args="--skip-tls-verify --build-arg A=val"
    :param force_build: Force building the image, even when no changes were made
    """
    engine, function = _get_engine_and_function(function, project_object)
    if function.kind in mlrun.runtimes.RuntimeKinds.nuclio_runtimes():
        raise mlrun.errors.MLRunInvalidArgumentError(
            "Cannot build use deploy_function()"
        )
    if engine == mlrun.common.schemas.workflow.EngineType.KFP:
        if overwrite_build_params:
            function.spec.build.commands = None
        if requirements or requirements_file:
            function.with_requirements(
                requirements=requirements,
                requirements_file=requirements_file,
                overwrite=True,
            )
        if commands:
            function.with_commands(commands)
        return function.deploy_step(
            image=image,
            base_image=base_image,
            commands=commands,
            secret_name=secret_name,
            with_mlrun=with_mlrun,
            skip_deployed=skip_deployed,
        )
    else:
        function.build_config(
            image=image,
            base_image=base_image,
            commands=commands,
            secret=secret_name,
            requirements=requirements,
            requirements_file=requirements_file,
            overwrite=overwrite_build_params,
            extra_args=extra_args,
        )
        ready = function.deploy(
            watch=True,
            with_mlrun=with_mlrun,
            skip_deployed=skip_deployed,
            mlrun_version_specifier=mlrun_version_specifier,
            builder_env=builder_env,
            force_build=force_build,
        )
        # return object with the same outputs as the KFP op (allow using the same pipeline)
        return BuildStatus(ready, {"image": function.spec.image}, function=function)


class DeployStatus:
    """returned status from deploy operation"""

    def __init__(self, state, outputs=None, function=None):
        self.state = state
        self.outputs = outputs or {}
        self.function = function

    def after(self, step):
        """nil function, for KFP compatibility"""
        return self

    def __repr__(self):
        return f"DeployStatus(state={self.state}, outputs={self.outputs})"


def deploy_function(
    function: Union[str, mlrun.runtimes.BaseRuntime],
    models: Optional[list] = None,
    env: Optional[dict] = None,
    tag: Optional[str] = None,
    verbose: Optional[bool] = None,
    builder_env: Optional[dict] = None,
    project_object=None,
    mock: Optional[bool] = None,
) -> Union[DeployStatus, mlrun_pipelines.models.PipelineNodeWrapper]:
    """deploy real-time (nuclio based) functions

    :param function:   name of the function (in the project) or function object
    :param models:     list of model items
    :param env:        dict of extra environment variables
    :param tag:        extra version tag
    :param verbose:    add verbose prints/logs
    :param builder_env: env vars dict for source archive config/credentials e.g. builder_env={"GIT_TOKEN": token}
    :param mock:       deploy mock server vs a real Nuclio function (for local simulations)
    :param project_object:  override the project object to use, will default to the project set in the runtime context.
    """
    engine, function = _get_engine_and_function(function, project_object)
    if function.kind not in mlrun.runtimes.RuntimeKinds.nuclio_runtimes():
        raise mlrun.errors.MLRunInvalidArgumentError(
            "deploy is used with real-time functions, for other kinds use build_function()"
        )
    if engine == mlrun.common.schemas.workflow.EngineType.KFP:
        return function.deploy_step(models=models, env=env, tag=tag, verbose=verbose)
    else:
        if env:
            function.set_envs(env)
        if models:
            for model_args in models:
                function.add_model(**model_args)

        mock = mlrun.mlconf.use_nuclio_mock(mock)
        function._set_as_mock(mock)
        if mock:
            # make sure the latest ver is saved in the DB (same as in function.deploy())
            function.save()
            return DeployStatus(
                state="ready",
                outputs={"endpoint": "Mock", "name": function.metadata.name},
                function=function,
            )

        address = function.deploy(tag=tag, verbose=verbose, builder_env=builder_env)
        # return object with the same outputs as the KFP op (allow using the same pipeline)
        return DeployStatus(
            state=function.status.state,
            outputs={"endpoint": address, "name": function.status.nuclio_name},
            function=function,
        )



# Copyright 2024 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import socket
from typing import Any, Optional, Protocol, cast

import nuclio.request
import numpy as np
import pandas as pd

import mlrun.common.constants as mlrun_constants
import mlrun.common.schemas.model_monitoring.constants as mm_constants
import mlrun.errors
import mlrun.feature_store as fstore
import mlrun.feature_store.feature_set as fs
import mlrun.serving
import mlrun.utils
from mlrun.artifacts import Artifact, DatasetArtifact, ModelArtifact, get_model
from mlrun.common.model_monitoring.helpers import FeatureStats
from mlrun.common.schemas import ModelEndpoint
from mlrun.model_monitoring.helpers import calculate_inputs_statistics


class _ArtifactsLogger(Protocol):
    """
    Classes that implement this protocol are :code:`MlrunProject` and :code:`MLClientCtx`.
    """

    def log_artifact(self, *args, **kwargs) -> Artifact: ...
    def log_dataset(self, *args, **kwargs) -> DatasetArtifact: ...


class MonitoringApplicationContext:
    _logger_name = "monitoring-application"

    def __init__(
        self,
        *,
        application_name: str,
        event: dict[str, Any],
        project: "mlrun.MlrunProject",
        artifacts_logger: _ArtifactsLogger,
        logger: mlrun.utils.Logger,
        nuclio_logger: nuclio.request.Logger,
        model_endpoint_dict: Optional[dict[str, ModelEndpoint]] = None,
        sample_df: Optional[pd.DataFrame] = None,
        feature_stats: Optional[FeatureStats] = None,
        feature_sets_dict: Optional[dict[str, fs.FeatureSet]] = None,
    ) -> None:
        """
        The :code:`MonitoringApplicationContext` object holds all the relevant information for the
        model monitoring application, and can be used for logging artifacts and messages.
        The monitoring context has the following attributes:

        :param application_name:        (str) The model monitoring application name.
        :param project:                 (:py:class:`~mlrun.projects.MlrunProject`) The current MLRun project object.
        :param project_name:            (str) The project name.
        :param logger:                  (:py:class:`~mlrun.utils.Logger`) MLRun logger.
        :param nuclio_logger:           (nuclio.request.Logger) Nuclio logger.
        :param sample_df_stats:         (FeatureStats) The new sample distribution dictionary.
        :param feature_stats:           (FeatureStats) The train sample distribution dictionary.
        :param sample_df:               (pd.DataFrame) The new sample DataFrame.
        :param start_infer_time:        (pd.Timestamp) Start time of the monitoring schedule.
        :param end_infer_time:          (pd.Timestamp) End time of the monitoring schedule.
        :param endpoint_id:             (str) ID of the monitored model endpoint
        :param feature_set:              (FeatureSet) the model endpoint feature set
        :param endpoint_name:           (str) Name of the monitored model endpoint
        :param output_stream_uri:       (str) URI of the output stream for results
        :param model_endpoint:          (ModelEndpoint) The model endpoint object.
        :param feature_names:           (list[str]) List of models feature names.
        :param label_names:             (list[str]) List of models label names.
        :param model:                   (tuple[str, ModelArtifact, dict]) The model file, model spec object,
                                        and a list of extra data items.
        """
        self.application_name = application_name

        self.project = project
        self.project_name = project.name

        self._artifacts_logger = artifacts_logger

        # MLRun Logger
        self.logger = logger
        # Nuclio logger - `nuclio.request.Logger`.
        # Note: this logger accepts keyword arguments only in its `_with` methods, e.g. `info_with`.
        self.nuclio_logger = nuclio_logger

        # event data
        self.start_infer_time = pd.Timestamp(
            cast(str, event.get(mm_constants.ApplicationEvent.START_INFER_TIME))
        )
        self.end_infer_time = pd.Timestamp(
            cast(str, event.get(mm_constants.ApplicationEvent.END_INFER_TIME))
        )
        self.endpoint_id = cast(
            str, event.get(mm_constants.ApplicationEvent.ENDPOINT_ID)
        )
        self.endpoint_name = cast(
            str, event.get(mm_constants.ApplicationEvent.ENDPOINT_NAME)
        )

        self._feature_stats: Optional[FeatureStats] = feature_stats
        self._sample_df_stats: Optional[FeatureStats] = None

        # Default labels for the artifacts
        self._default_labels = self._get_default_labels()

        # Persistent data - fetched when needed
        self._sample_df: Optional[pd.DataFrame] = sample_df
        self._model_endpoint: Optional[ModelEndpoint] = (
            model_endpoint_dict.get(self.endpoint_id) if model_endpoint_dict else None
        )
        self._feature_set: Optional[fs.FeatureSet] = (
            feature_sets_dict.get(self.endpoint_id) if feature_sets_dict else None
        )
        store, _, _ = mlrun.store_manager.get_or_create_store(
            mlrun.mlconf.artifact_path
        )
        self.storage_options = store.get_storage_options()

    @classmethod
    def _from_ml_ctx(
        cls,
        context: "mlrun.MLClientCtx",
        *,
        project: Optional["mlrun.MlrunProject"] = None,
        application_name: str,
        event: dict[str, Any],
        model_endpoint_dict: Optional[dict[str, ModelEndpoint]] = None,
        sample_df: Optional[pd.DataFrame] = None,
        feature_stats: Optional[FeatureStats] = None,
    ) -> "MonitoringApplicationContext":
        project = project or context.get_project_object()
        if not project:
            raise mlrun.errors.MLRunValueError("Could not load project from context")
        logger = context.logger
        artifacts_logger = context
        nuclio_logger = nuclio.request.Logger(
            level=mlrun.mlconf.log_level, name=cls._logger_name
        )
        return cls(
            application_name=application_name,
            event=event,
            model_endpoint_dict=model_endpoint_dict,
            project=project,
            logger=logger,
            nuclio_logger=nuclio_logger,
            artifacts_logger=artifacts_logger,
            sample_df=sample_df,
            feature_stats=feature_stats,
        )

    @classmethod
    def _from_graph_ctx(
        cls,
        graph_context: mlrun.serving.GraphContext,
        *,
        application_name: str,
        event: dict[str, Any],
        model_endpoint_dict: Optional[dict[str, ModelEndpoint]] = None,
        sample_df: Optional[pd.DataFrame] = None,
        feature_stats: Optional[FeatureStats] = None,
        feature_sets_dict: Optional[dict[str, fs.FeatureSet]] = None,
    ) -> "MonitoringApplicationContext":
        nuclio_logger = graph_context.logger
        artifacts_logger = graph_context.project_obj
        logger = mlrun.utils.create_logger(
            level=mlrun.mlconf.log_level,
            formatter_kind=mlrun.mlconf.log_formatter,
            name=cls._logger_name,
        )
        return cls(
            application_name=application_name,
            event=event,
            project=graph_context.project_obj,
            model_endpoint_dict=model_endpoint_dict,
            logger=logger,
            nuclio_logger=nuclio_logger,
            artifacts_logger=artifacts_logger,
            sample_df=sample_df,
            feature_stats=feature_stats,
            feature_sets_dict=feature_sets_dict,
        )

    def _get_default_labels(self) -> dict[str, str]:
        labels = {
            mlrun_constants.MLRunInternalLabels.runner_pod: socket.gethostname(),
            mlrun_constants.MLRunInternalLabels.producer_type: "model-monitoring-app",
            mlrun_constants.MLRunInternalLabels.app_name: self.application_name,
            mlrun_constants.MLRunInternalLabels.endpoint_id: self.endpoint_id,
            mlrun_constants.MLRunInternalLabels.endpoint_name: self.endpoint_name,
        }
        return {key: value for key, value in labels.items() if value is not None}

    def _add_default_labels(self, labels: Optional[dict[str, str]]) -> dict[str, str]:
        """Add the default labels to logged artifacts labels"""
        return (labels or {}) | self._default_labels

    @property
    def sample_df(self) -> pd.DataFrame:
        if self._sample_df is None:
            if (
                self.endpoint_name is None
                or self.endpoint_id is None
                or pd.isnull(self.start_infer_time)
                or pd.isnull(self.end_infer_time)
            ):
                raise mlrun.errors.MLRunValueError(
                    "You have tried to access `monitoring_context.sample_df`, but have not provided it directly "
                    "through `sample_data`, nor have you provided the model endpoint's name, ID, and the start and "
                    f"end times: `endpoint_name`={self.endpoint_name}, `endpoint_uid`={self.endpoint_id}, "
                    f"`start`={self.start_infer_time}, and `end`={self.end_infer_time}. "
                    "You can either provide the sample dataframe directly, the model endpoint's details and times, "
                    "or adapt the application's logic to not access the sample dataframe."
                )
            df = self.feature_set.to_dataframe(
                start_time=self.start_infer_time,
                end_time=self.end_infer_time,
                time_column=mm_constants.EventFieldType.TIMESTAMP,
                storage_options=self.storage_options,
            )
            self._sample_df = df.reset_index(drop=True)
        return self._sample_df

    @property
    def model_endpoint(self) -> ModelEndpoint:
        if not self._model_endpoint:
            if self.endpoint_name is None or self.endpoint_id is None:
                raise mlrun.errors.MLRunValueError(
                    "You have NOT provided the model endpoint's name and ID: "
                    f"`endpoint_name`={self.endpoint_name} and `endpoint_id`={self.endpoint_id}, "
                    "but you have tried to access `monitoring_context.model_endpoint` "
                    "directly or indirectly in your application. You can either provide them, "
                    "or adapt the application's logic to not access the model endpoint."
                )
            self._model_endpoint = mlrun.db.get_run_db().get_model_endpoint(
                name=self.endpoint_name,
                project=self.project_name,
                endpoint_id=self.endpoint_id,
                feature_analysis=True,
                tsdb_metrics=False,
            )
        return self._model_endpoint

    @property
    def feature_set(self) -> fs.FeatureSet:
        if not self._feature_set and self.model_endpoint:
            self._feature_set = fstore.get_feature_set(
                self.model_endpoint.spec.monitoring_feature_set_uri
            )
        return self._feature_set

    @property
    def feature_stats(self) -> FeatureStats:
        if not self._feature_stats:
            self._feature_stats = self.model_endpoint.spec.feature_stats
        return self._feature_stats

    @property
    def sample_df_stats(self) -> FeatureStats:
        """statistics of the sample dataframe"""
        if not self._sample_df_stats:
            self._sample_df_stats = calculate_inputs_statistics(
                self.feature_stats, self.sample_df
            )
        return self._sample_df_stats

    @property
    def feature_names(self) -> list[str]:
        """The feature names of the model"""
        return self.model_endpoint.spec.feature_names

    @property
    def label_names(self) -> list[str]:
        """The label names of the model"""
        return self.model_endpoint.spec.label_names

    @property
    def model(self) -> tuple[str, ModelArtifact, dict]:
        """The model file, model spec object, and a list of extra data items"""
        return get_model(self.model_endpoint.spec.model_uri)

    @staticmethod
    def dict_to_histogram(histogram_dict: FeatureStats) -> pd.DataFrame:
        """
        Convert histogram dictionary to pandas DataFrame with feature histograms as columns

        :param histogram_dict: Histogram dictionary

        :returns: Histogram dataframe
        """

        # Create a dictionary with feature histograms as values
        histograms = {}
        for feature, stats in histogram_dict.items():
            if "hist" in stats:
                # Normalize to probability distribution of each feature
                histograms[feature] = np.array(stats["hist"][0]) / stats["count"]

        # Convert the dictionary to pandas DataFrame
        histograms = pd.DataFrame(histograms)

        return histograms

    def log_artifact(
        self,
        item,
        body=None,
        tag: str = "",
        local_path: str = "",
        artifact_path: Optional[str] = None,
        format: Optional[str] = None,
        upload: Optional[bool] = None,
        labels: Optional[dict[str, str]] = None,
        target_path: Optional[str] = None,
        unique_per_endpoint: bool = True,
        **kwargs,
    ) -> Artifact:
        """
        Log an artifact.

        .. caution::

            Logging artifacts in every model monitoring window may cause scale issues.
            This method should be called on special occasions only.

        See :func:`~mlrun.projects.MlrunProject.log_artifact` for the full documentation, except for one
        new argument:

        :param unique_per_endpoint: by default ``True``, we will log different artifact for each model endpoint,
                                    set to ``False`` without changing item key will cause artifact override.
        """
        labels = self._add_default_labels(labels)
        # By default, we want to log different artifact for each model endpoint
        endpoint_id = labels.get(mlrun_constants.MLRunInternalLabels.endpoint_id, "")
        if unique_per_endpoint and isinstance(item, str):
            item = f"{item}-{endpoint_id}" if endpoint_id else item
        elif unique_per_endpoint:  # isinstance(item, Artifact) is True
            item.key = f"{item.key}-{endpoint_id}" if endpoint_id else item.key

        return self._artifacts_logger.log_artifact(
            item,
            body=body,
            tag=tag,
            local_path=local_path,
            artifact_path=artifact_path,
            format=format,
            upload=upload,
            labels=labels,
            target_path=target_path,
            **kwargs,
        )

    def log_dataset(
        self,
        key,
        df,
        tag="",
        local_path=None,
        artifact_path=None,
        upload=None,
        labels=None,
        format="",
        preview=None,
        stats=None,
        target_path="",
        extra_data=None,
        label_column: Optional[str] = None,
        unique_per_endpoint: bool = True,
        **kwargs,
    ) -> DatasetArtifact:
        """
        Log a dataset artifact.

        .. caution::

            Logging datasets in every model monitoring window may cause scale issues.
            This method should be called on special occasions only.

        See :func:`~mlrun.projects.MlrunProject.log_dataset` for the full documentation, except for one
        new argument:

        :param unique_per_endpoint: by default ``True``, we will log different artifact for each model endpoint,
                                    set to ``False`` without changing item key will cause artifact override.
        """
        labels = self._add_default_labels(labels)
        # By default, we want to log different artifact for each model endpoint
        endpoint_id = labels.get(mlrun_constants.MLRunInternalLabels.endpoint_id, "")
        if unique_per_endpoint and isinstance(key, str):
            key = f"{key}-{endpoint_id}" if endpoint_id else key

        return self._artifacts_logger.log_dataset(
            key,
            df,
            tag=tag,
            local_path=local_path,
            artifact_path=artifact_path,
            upload=upload,
            labels=labels,
            format=format,
            preview=preview,
            stats=stats,
            target_path=target_path,
            extra_data=extra_data,
            label_column=label_column,
            **kwargs,
        )



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Optional, Union

import numpy as np
from tensorflow import keras

import mlrun
from mlrun.serving.v2_serving import V2ModelServer

from .model_handler import TFKerasModelHandler


class TFKerasModelServer(V2ModelServer):
    """
    Tensorflow.keras Model serving class, inheriting the V2ModelServer class for being initialized automatically by the
    model server and be able to run locally as part of a nuclio serverless function, or as part of a real-time pipeline.
    """

    def __init__(
        self,
        context: mlrun.MLClientCtx = None,
        name: Optional[str] = None,
        model: keras.Model = None,
        model_path: Optional[str] = None,
        model_name: Optional[str] = None,
        modules_map: Optional[
            Union[dict[str, Union[None, str, list[str]]], str]
        ] = None,
        custom_objects_map: Optional[
            Union[dict[str, Union[str, list[str]]], str]
        ] = None,
        custom_objects_directory: Optional[str] = None,
        model_format: str = TFKerasModelHandler.ModelFormats.SAVED_MODEL,
        to_list: bool = False,
        protocol: Optional[str] = None,
        **class_args,
    ):
        """
        Initialize a serving class for a tf.keras model.

        :param context:                  For internal use (passed in init).
        :param name:                     The model name to be served.
        :param model:                    Model to handle or None in case a loading parameters were supplied.
        :param model_path:               Path to the model's directory to load it from. The model files must start with
                                         the given model name and the directory must contain based on the given model
                                         formats:
                                         * SavedModel - A zip file 'model_name.zip' or a directory named 'model_name'.
                                         * H5 - A h5 file 'model_name.h5'.
                                         * Architecture and weights - The json file 'model_name.json' and h5 weight file
                                           'model_name.h5'.
                                         The model path can be also passed as a model object path in the following
                                         format: 'store://models/<PROJECT_NAME>/<MODEL_NAME>:<VERSION>'.
        :param model_name:               The model name for saving and logging the model:
                                         * Mandatory for loading the model from a local path.
                                         * If given a logged model (store model path) it will be read from the artifact.
                                         * If given a loaded model object and the model name is None, the name will be
                                           set to the model's object name / class.
        :param modules_map:              A dictionary of all the modules required for loading the model. Each key
                                         is a path to a module and its value is the object name to import from it. All
                                         the modules will be imported globally. If multiple objects needed to be
                                         imported from the same module a list can be given. The map can be passed as a
                                         path to a json file as well. For example:
                                         {
                                             "module1": None,  # => import module1
                                             "module2": ["func1", "func2"],  # => from module2 import func1, func2
                                             "module3.sub_module": "func3",  # => from module3.sub_module import func3
                                         }
                                         If the model path given is of a store object, the modules map will be read from
                                         the logged modules map artifact of the model.
        :param custom_objects_map:       A dictionary of all the custom objects required for loading the model. Each key
                                         is a path to a python file and its value is the custom object name to import
                                         from it. If multiple objects needed to be imported from the same py file a list
                                         can be given. The map can be passed as a path to a json file as well. For
                                         example:
                                         {
                                             "/.../custom_optimizer.py": "optimizer",
                                             "/.../custom_layers.py": ["layer1", "layer2"]
                                         }
                                         All the paths will be accessed from the given 'custom_objects_directory',
                                         meaning each py file will be read from 'custom_objects_directory/<MAP VALUE>'.
                                         If the model path given is of a store object, the custom objects map will be
                                         read from the logged custom object map artifact of the model.
                                         Notice: The custom objects will be imported in the order they came in this
                                         dictionary (or json). If a custom object is depended on another, make sure to
                                         put it below the one it relies on.
        :param custom_objects_directory: Path to the directory with all the python files required for the custom
                                         objects. Can be passed as a zip file as well (will be extracted during the run
                                         before loading the model). If the model path given is of a store object, the
                                         custom objects files will be read from the logged custom object artifact of the
                                         model.
        :param model_format:             The format used to save the model. One of the members of the
                                         TFKerasModelHandler.ModelFormats class. Default: SavedModel.
        :param to_list:                  Whether to return a list instead of a numpy.ndarray. Default: False.
        :param protocol:                 -
        :param class_args:               -
        """
        super().__init__(
            context=context,
            name=name,
            model_path=model_path,
            model=model,
            protocol=protocol,
            **class_args,
        )

        # Store the model handler attributes:
        self.model_name = model_name
        self.modules_map = modules_map
        self.custom_objects_map = custom_objects_map
        self.custom_objects_directory = custom_objects_directory
        self.model_format = model_format

        # Store additional configurations:
        self.to_list = to_list

        # Set up a model handler:
        self._model_handler: TFKerasModelHandler = None

    def load(self):
        """
        Use the model handler to load the model.
        """
        # Initialize the model handler:
        self._model_handler = TFKerasModelHandler(
            model=self.model,
            model_path=self.model_path,
            model_name=self.model_name,
            modules_map=self.modules_map,
            custom_objects_map=self.custom_objects_map,
            custom_objects_directory=self.custom_objects_directory,
            model_format=self.model_format,
            context=self.context,
        )

        # Load the model:
        if self._model_handler.model is None:
            self._model_handler.load()
        self.model = self._model_handler.model

    def predict(self, request: dict[str, Any]) -> Union[np.ndarray, list]:
        """
        Infer the inputs through the model using 'keras.Model.predict' and return its output. The inferred data will be
        read from the "inputs" key of the request.

        :param request: The request to the model. The input to the model will be read from the "inputs" key.

        :return: The 'keras.Model.predict' returned output on the given inputs. If 'to_list' was set to True in
                 initialization, a list will be returned instead of a numpy.ndarray.
        """
        # Get the inputs:
        inputs = request["inputs"]

        # Predict:
        prediction = self.model.predict(inputs)

        # Return as list if required:
        return prediction if not self.to_list else prediction.tolist()

    def explain(self, request: dict[str, Any]) -> str:
        """
        Return a string explaining what model is being serve in this serving function and the function name.

        :param request: A given request.

        :return: Explanation string.
        """
        return f"The '{self.model.name}' model serving function named '{self.name}'"



# Copyright 2024 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime, timedelta

import mlrun.common.types


class PartitionInterval(mlrun.common.types.StrEnum):
    DAY = "DAY"
    MONTH = "MONTH"
    YEARWEEK = "YEARWEEK"

    @classmethod
    def is_valid(cls, value: str) -> bool:
        return value in cls._value2member_map_

    @classmethod
    def valid_intervals(cls) -> list:
        return list(cls._value2member_map_.keys())

    def as_duration(self) -> timedelta:
        """
        Convert the partition interval to a duration-like timedelta.

        Returns:
            timedelta: A duration representing the partition interval.
        """
        if self == PartitionInterval.DAY:
            return timedelta(days=1)
        elif self == PartitionInterval.MONTH:
            # Approximate a month as 30 days
            return timedelta(days=30)
        elif self == PartitionInterval.YEARWEEK:
            return timedelta(weeks=1)
        else:
            raise ValueError(f"Unsupported PartitionInterval: {self}")

    @classmethod
    def from_expression(cls, partition_expression: str):
        """
        Returns the corresponding PartitionInterval for a given partition expression,
        or None if the function is not mapped.

        :param partition_expression: The partition expression to map to an interval.
        :return: PartitionInterval corresponding to the expression, or `month` if no match is found.
        """

        # Match the provided function string to the correct interval
        partition_expression = partition_expression.upper()
        if "YEARWEEK" in partition_expression:
            return cls.YEARWEEK
        elif "DAYOFMONTH" in partition_expression:
            return cls.DAY
        else:
            return cls.MONTH

    def get_partition_info(
        self,
        start_datetime: datetime,
        partition_number: int = 1,
    ) -> list[tuple[str, str]]:
        """
        Generates partition details for a specified number of partitions starting from a given datetime.

        :param start_datetime: The starting datetime used for generating partition details.
        :param partition_number: The number of partitions to generate details for.

        :return: A list of tuples:
            - partition_name: The name for the partition.
            - partition_value: The "LESS THAN" value for the next partition boundary.
        """
        partitioning_information_list = []
        current_datetime = start_datetime

        for _ in range(partition_number):
            partition_name = f"p{self.get_partition_name(current_datetime)}"
            partition_boundary_date = self.get_next_partition_time(current_datetime)
            partition_value = self.get_partition_name(partition_boundary_date)
            partitioning_information_list.append((partition_name, partition_value))

            # Move to the next interval
            current_datetime = partition_boundary_date

        return partitioning_information_list

    def get_next_partition_time(self, current_datetime: datetime) -> datetime:
        """
        Calculates the next partition boundary time based on the specified partition interval.
        :param current_datetime: The current datetime from which the next interval is calculated.

        :return: A datetime object representing the start of the next partition interval.
            - If the interval is DAY, it advances by one day.
            - If the interval is MONTH, it advances to the first day of the next month.
            - If the interval is YEARWEEK, it advances by one week.
        """
        if self == PartitionInterval.DAY:
            return current_datetime + timedelta(days=1)
        elif self == PartitionInterval.MONTH:
            return (current_datetime.replace(day=1) + timedelta(days=32)).replace(day=1)
        elif self == PartitionInterval.YEARWEEK:
            return current_datetime + timedelta(weeks=1)
        else:
            raise ValueError(f"Unsupported PartitionInterval: {self}")

    def get_partition_name(self, current_datetime: datetime) -> str:
        if self == PartitionInterval.DAY:
            return current_datetime.strftime("%Y%m%d")
        elif self == PartitionInterval.MONTH:
            return current_datetime.strftime("%Y%m")
        elif self == PartitionInterval.YEARWEEK:
            year, week, _ = current_datetime.isocalendar()
            return f"{year}{week:02d}"
        else:
            raise ValueError(f"Unsupported PartitionInterval: {self}")

    def get_partition_expression(self, column_name: str):
        if self == PartitionInterval.YEARWEEK:
            return f"YEARWEEK({column_name}, 1)"
        elif self == PartitionInterval.DAY:
            # generates value in format %Y%m%d in mysql
            # mysql query example: `select YEAR(NOW())*10000 + MONTH(NOW())*100 + DAY(NOW());`
            return f"YEAR({column_name}) * 10000 + MONTH({column_name}) * 100 + DAY({column_name})"
        elif self == PartitionInterval.MONTH:
            # generates value in format %Y%m in mysql
            # mysql query example: `select YEAR(NOW())*100 + MONTH(NOW());`
            return f"YEAR({column_name}) * 100 + MONTH({column_name})"
        else:
            raise ValueError(f"Unsupported PartitionInterval: {self}")

    def get_number_of_partitions(self, days: int) -> int:
        # Calculate the number partitions based on given number of days
        if self == PartitionInterval.DAY:
            return days
        elif self == PartitionInterval.MONTH:
            # Average number days in a month is 30.44
            return int(days / 30.44)
        elif self == PartitionInterval.YEARWEEK:
            return int(days / 7)
        else:
            raise ValueError(f"Unsupported PartitionInterval: {self}")



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .data_types import (
    InferOptions,
    ValueType,
    pd_schema_to_value_type,
    python_type_to_value_type,
)
from .infer import DFDataInfer


class BaseDataInfer:
    infer_schema = None
    get_preview = None
    get_stats = None


def is_spark_dataframe(df) -> bool:
    return "rdd" in dir(df)


def get_infer_interface(df) -> BaseDataInfer:
    if is_spark_dataframe(df):
        from .spark import SparkDataInfer

        return SparkDataInfer
    return DFDataInfer



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typing

import pydantic.v1

import mlrun.common.types


class ResourceSpec(pydantic.v1.BaseModel):
    cpu: typing.Optional[str]
    memory: typing.Optional[str]
    gpu: typing.Optional[str]


class Resources(pydantic.v1.BaseModel):
    requests: ResourceSpec = ResourceSpec()
    limits: ResourceSpec = ResourceSpec()


class NodeSelectorOperator(mlrun.common.types.StrEnum):
    """
    A node selector operator is the set of operators that can be used in a node selector requirement
    https://github.com/kubernetes/api/blob/b754a94214be15ffc8d648f9fe6481857f1fc2fe/core/v1/types.go#L2765
    """

    node_selector_op_in = "In"
    node_selector_op_not_in = "NotIn"
    node_selector_op_exists = "Exists"
    node_selector_op_does_not_exist = "DoesNotExist"
    node_selector_op_gt = "Gt"
    node_selector_op_lt = "Lt"



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re
from subprocess import run
from typing import Optional

import kubernetes.client

import mlrun.errors
from mlrun.config import config
from mlrun.runtimes.mounts import mount_v3io, mount_v3iod

from .kubejob import KubejobRuntime
from .pod import KubeResourceSpec


class RemoteSparkSpec(KubeResourceSpec):
    _dict_fields = KubeResourceSpec._dict_fields + ["provider"]

    def __init__(
        self,
        command=None,
        args=None,
        image=None,
        mode=None,
        volumes=None,
        volume_mounts=None,
        env=None,
        resources=None,
        default_handler=None,
        entry_points=None,
        description=None,
        workdir=None,
        replicas=None,
        image_pull_policy=None,
        service_account=None,
        build=None,
        image_pull_secret=None,
        provider=None,
        node_name=None,
        node_selector=None,
        affinity=None,
        priority_class_name=None,
        disable_auto_mount=False,
        pythonpath=None,
        tolerations=None,
        preemption_mode=None,
        security_context=None,
        state_thresholds=None,
        serving_spec=None,
        graph=None,
        parameters=None,
        track_models=None,
    ):
        super().__init__(
            command=command,
            args=args,
            image=image,
            mode=mode,
            volumes=volumes,
            volume_mounts=volume_mounts,
            env=env,
            resources=resources,
            default_handler=default_handler,
            entry_points=entry_points,
            description=description,
            workdir=workdir,
            replicas=replicas,
            image_pull_policy=image_pull_policy,
            service_account=service_account,
            build=build,
            image_pull_secret=image_pull_secret,
            node_name=node_name,
            node_selector=node_selector,
            affinity=affinity,
            priority_class_name=priority_class_name,
            disable_auto_mount=disable_auto_mount,
            pythonpath=pythonpath,
            tolerations=tolerations,
            preemption_mode=preemption_mode,
            security_context=security_context,
            state_thresholds=state_thresholds,
            serving_spec=serving_spec,
            graph=graph,
            parameters=parameters,
            track_models=track_models,
        )
        self.provider = provider


class RemoteSparkProviders:
    iguazio = "iguazio"


class RemoteSparkRuntime(KubejobRuntime):
    kind = "remote-spark"
    default_image = ".remote-spark-default-image"

    @classmethod
    def deploy_default_image(cls):
        if not mlrun.get_current_project(silent=True):
            raise mlrun.errors.MLRunMissingProjectError(
                "An active project is required to run deploy_default_image(). "
                "This can be set by calling get_or_create_project(), load_project(), or new_project()."
            )

        sj = mlrun.new_function(
            kind="remote-spark", name="remote-spark-default-image-deploy-temp"
        )
        sj.spec.build.image = cls.default_image
        sj.with_spark_service(spark_service="dummy-spark")
        sj.deploy()
        mlrun.get_run_db().delete_function(name=sj.metadata.name)

    def is_deployed(self):
        if (
            not self.spec.build.source
            and not self.spec.build.commands
            and not self.spec.build.extra
        ):
            return True
        return super().is_deployed()

    @property
    def spec(self) -> RemoteSparkSpec:
        return self._spec

    @spec.setter
    def spec(self, spec):
        self._spec = self._verify_dict(spec, "spec", RemoteSparkSpec)

    def with_spark_service(
        self,
        spark_service,
        provider=RemoteSparkProviders.iguazio,
        with_v3io_mount=True,
    ):
        """Attach spark service to function"""
        self.spec.provider = provider
        if provider == RemoteSparkProviders.iguazio:
            self.spec.env.append(
                {"name": "MLRUN_SPARK_CLIENT_IGZ_SPARK", "value": "true"}
            )
            if with_v3io_mount:
                self.apply(mount_v3io())
            self.apply(
                mount_v3iod(
                    namespace=config.namespace,
                    v3io_config_configmap=spark_service + "-submit",
                )
            )

    def with_security_context(
        self, security_context: kubernetes.client.V1SecurityContext
    ):
        """
        With security context is not supported for spark runtime.
        Driver / Executor processes run with uid / gid 1000 as long as security context is not defined.
        If in the future we want to support setting security context it will work only from spark version 3.2 onwards.
        """
        raise mlrun.errors.MLRunInvalidArgumentTypeError(
            "with_security_context is not supported with remote spark"
        )

    @property
    def _resolve_default_base_image(self):
        if (
            self.spec.provider == RemoteSparkProviders.iguazio
            and config.spark_app_image
            and config.spark_app_image_tag
        ):
            app_image = re.sub("spark-app", "shell", config.spark_app_image)
            # this is temporary until we get the image name from external config
            return app_image + ":" + config.spark_app_image_tag
        return None

    def deploy(
        self,
        watch=True,
        with_mlrun=None,
        skip_deployed=False,
        is_kfp=False,
        mlrun_version_specifier=None,
        builder_env: Optional[dict] = None,
        show_on_failure: bool = False,
        force_build: bool = False,
    ):
        """deploy function, build container with dependencies

        :param watch:                   wait for the deploy to complete (and print build logs)
        :param with_mlrun:              add the current mlrun package to the container build
        :param skip_deployed:           skip the build if we already have an image for the function
        :param is_kfp:                  deploy as part of a kfp pipeline
        :param mlrun_version_specifier: which mlrun package version to include (if not current)
        :param builder_env:             Kaniko builder pod env vars dict (for config/credentials)
                                        e.g. builder_env={"GIT_TOKEN": token}
        :param show_on_failure:         show logs only in case of build failure
        :param force_build:             force building the image, even when no changes were made

        :return True if the function is ready (deployed)
        """
        # connect will populate the config from the server config
        if not self.spec.build.base_image:
            self.spec.build.base_image = self._resolve_default_base_image
        return super().deploy(
            watch=watch,
            with_mlrun=with_mlrun,
            skip_deployed=skip_deployed,
            is_kfp=is_kfp,
            mlrun_version_specifier=mlrun_version_specifier,
            builder_env=builder_env,
            show_on_failure=show_on_failure,
            force_build=force_build,
        )


def igz_spark_pre_hook():
    run(["/bin/bash", "/etc/config/v3io/spark-job-init.sh"])



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import hashlib
import os
import pathlib
import tempfile
import typing
import zipfile

import yaml

import mlrun
import mlrun.artifacts
import mlrun.datastore
import mlrun.errors

from ..model import ModelObj
from ..utils import (
    StorePrefix,
    calculate_local_file_hash,
    generate_artifact_uri,
    is_relative_path,
)


class ArtifactMetadata(ModelObj):
    _dict_fields = [
        "key",
        "project",
        "iter",
        "tree",
        "description",
        "hash",
        "tag",
        "uid",
    ]
    _extra_fields = ["updated", "labels"]

    def __init__(
        self,
        key=None,
        project=None,
        iter=None,
        tree=None,
        description=None,
        hash=None,
        tag=None,
        uid=None,
    ):
        self.key = key
        self.project = project
        self.iter = iter
        self.tree = tree
        self.description = description
        self.hash = hash
        self.labels = {}
        self.updated = None
        self.tag = tag  # temp store of the tag
        self.uid = uid

    def base_dict(self):
        return super().to_dict()

    def to_dict(
        self,
        fields: typing.Optional[list] = None,
        exclude: typing.Optional[list] = None,
        strip: bool = False,
    ):
        """return long dict form of the artifact"""
        return super().to_dict(
            self._dict_fields + self._extra_fields, exclude=exclude, strip=strip
        )

    @classmethod
    def from_dict(
        cls, struct=None, fields=None, deprecated_fields: typing.Optional[dict] = None
    ):
        fields = fields or cls._dict_fields + cls._extra_fields
        return super().from_dict(
            struct, fields=fields, deprecated_fields=deprecated_fields
        )


class ArtifactSpec(ModelObj):
    _dict_fields = [
        "src_path",
        "target_path",
        "viewer",
        "inline",
        "format",
        "size",
        "db_key",
        "extra_data",
        "unpackaging_instructions",
        "producer",
        "parent_uri",
        "has_children",
    ]

    _extra_fields = ["annotations", "sources", "license", "encoding"]
    _exclude_fields_from_uid_hash = [
        # if the artifact is first created, it will not have a db_key,
        # exclude it so further updates of the artifacts will have the same hash
        "db_key",
        "extra_data",
    ]

    def __init__(
        self,
        src_path=None,
        target_path=None,
        viewer=None,
        is_inline=False,
        format=None,
        size=None,
        db_key=None,
        extra_data=None,
        body=None,
        unpackaging_instructions: typing.Optional[dict] = None,
        parent_uri: typing.Optional[str] = None,
    ):
        self.src_path = src_path
        self.target_path = target_path
        self.viewer = viewer
        self._is_inline = is_inline
        self.format = format
        self.size = size
        self.db_key = db_key
        self.extra_data = extra_data or {}
        self.unpackaging_instructions = unpackaging_instructions
        self.parent_uri = parent_uri
        self.has_children = False

        self._body = body
        self.encoding = None
        self.annotations = None
        self.sources = []
        self.producer = None
        self.license = ""

    def base_dict(self):
        return super().to_dict()

    def to_dict(
        self,
        fields: typing.Optional[list] = None,
        exclude: typing.Optional[list] = None,
        strip: bool = False,
    ):
        """return long dict form of the artifact"""
        return super().to_dict(
            self._dict_fields + self._extra_fields, exclude=exclude, strip=strip
        )

    @classmethod
    def from_dict(
        cls, struct=None, fields=None, deprecated_fields: typing.Optional[dict] = None
    ):
        fields = fields or cls._dict_fields + cls._extra_fields
        return super().from_dict(
            struct, fields=fields, deprecated_fields=deprecated_fields
        )

    @property
    def inline(self):
        """inline data (body)"""

        if self._is_inline:
            return self.get_body()
        return None

    @inline.setter
    def inline(self, body):
        self._body = body
        if body:
            self._is_inline = True

    def get_body(self):
        """get the artifact body when inline"""
        return self._body


class ArtifactStatus(ModelObj):
    _dict_fields = ["state", "stats", "preview", "header_original_length"]

    def __init__(self):
        self.state = "created"
        self.stats = None
        self.preview = None
        self.header_original_length = None

    def base_dict(self):
        return super().to_dict()


class Artifact(ModelObj):
    kind = "artifact"
    _dict_fields = ["kind", "metadata", "spec", "status"]

    _store_prefix = StorePrefix.Artifact

    def __init__(
        self,
        key=None,
        body=None,
        viewer=None,
        is_inline=False,
        format=None,
        size=None,
        target_path=None,
        project=None,
        src_path: typing.Optional[str] = None,
        metadata: ArtifactMetadata = None,
        spec: ArtifactSpec = None,
    ):
        self._metadata = None
        self.metadata = metadata
        self._spec = None
        self.spec = spec

        self.metadata.key = key or self.metadata.key
        self.metadata.project = (
            project or mlrun.mlconf.active_project or self.metadata.project
        )
        self.spec.size = size or self.spec.size
        self.spec.target_path = target_path or self.spec.target_path
        self.spec.format = format or self.spec.format
        self.spec.viewer = viewer or self.spec.viewer
        self.spec.src_path = src_path or self.spec.src_path

        # temp flag to indicate if the source path is a temporary file (if True it will be deleted after upload)
        self._src_is_temp = False

        if body:
            self.spec._body = body
        self.spec._is_inline = is_inline or self.spec._is_inline

        self.status = ArtifactStatus()

    @property
    def metadata(self) -> ArtifactMetadata:
        return self._metadata

    @metadata.setter
    def metadata(self, metadata):
        self._metadata = self._verify_dict(metadata, "metadata", ArtifactMetadata)

    @property
    def spec(self) -> ArtifactSpec:
        return self._spec

    @spec.setter
    def spec(self, spec):
        self._spec = self._verify_dict(spec, "spec", ArtifactSpec)

    @property
    def status(self) -> ArtifactStatus:
        return self._status

    @status.setter
    def status(self, status):
        self._status = self._verify_dict(status, "status", ArtifactStatus)

    def _get_file_body(self):
        body = self.spec.get_body()
        if body:
            return body
        if self.src_path and os.path.isfile(self.src_path):
            with open(self.src_path, "rb") as fp:
                return fp.read()
        return mlrun.get_dataitem(self.get_target_path()).get()

    def export(self, target_path: str, with_extras=True):
        """save the artifact object into a yaml/json file or zip archive

        when the target path is a .yaml/.json file the artifact spec is saved into that file,
        when the target_path suffix is '.zip' the artifact spec, body and extra data items are
        packaged into a zip file. The archive target_path support DataItem urls for remote object storage
        (e.g. s3://<bucket>/<path>).

        :param target_path: path to store artifact .yaml/.json spec or .zip (spec with the content)
        :param with_extras: will include the extra_data items in the zip archive
        """
        if target_path.endswith(".yaml") or target_path.endswith(".yml"):
            mlrun.get_dataitem(target_path).put(self.to_yaml())

        elif target_path.endswith(".json"):
            mlrun.get_dataitem(target_path).put(self.to_json())

        elif target_path.endswith(".zip"):
            tmp_path = None
            if "://" in target_path:
                tmp_path = tempfile.NamedTemporaryFile(suffix=".zip", delete=False).name
            zipf = zipfile.ZipFile(tmp_path or target_path, "w")
            body = self._get_file_body()
            zipf.writestr("_body", body)
            extras = {}
            if with_extras:
                for k, item_path in self.extra_data.items():
                    if is_relative_path(item_path):
                        base_dir = self.src_path or ""
                        if not self.is_dir:
                            base_dir = os.path.dirname(base_dir)
                        item_path = os.path.join(base_dir, item_path).replace("\\", "/")
                    zipf.writestr(k, mlrun.get_dataitem(item_path).get())
                    extras[k] = k
            artifact = self.copy()
            artifact.extra_data = extras
            zipf.writestr("_spec.yaml", artifact.to_yaml())
            zipf.close()

            if tmp_path:
                mlrun.get_dataitem(target_path).upload(tmp_path)
                os.remove(tmp_path)
        else:
            raise ValueError("unsupported file suffix, use .yaml, .json, or .zip")

    def before_log(self):
        for key, item in self.spec.extra_data.items():
            if hasattr(item, "get_target_path") and item.get_target_path():
                self.spec.extra_data[key] = item.get_target_path()

    @property
    def is_dir(self):
        """this is a directory"""
        return False

    @property
    def uri(self):
        """return artifact uri (store://..)"""
        return self.get_store_url()

    def to_dataitem(self):
        """return a DataItem object (if available) representing the artifact content"""
        uri = self.get_store_url()
        if uri:
            return mlrun.get_dataitem(uri)

    def get_body(self):
        """get the artifact body when inline"""
        return self.spec.get_body()

    def get_target_path(self):
        """get the absolute target path for the artifact"""
        return self.spec.target_path

    def get_store_url(self, with_tag=True, project=None, with_tree=True):
        """get the artifact uri (store://..) with optional parameters"""
        tag = self.metadata.tag if with_tag else None
        tree = self.metadata.tree if with_tree else None

        uri = generate_artifact_uri(
            project or self.metadata.project,
            self.spec.db_key,
            iter=self.metadata.iter,
            tree=tree,
            tag=tag,
            uid=self.uid,
        )
        return mlrun.datastore.get_store_uri(self._store_prefix, uri)

    def base_dict(self):
        """return short dict form of the artifact"""
        struct = {"kind": self.kind}
        for field in ["metadata", "spec", "status"]:
            val = getattr(self, field, None)
            if val:
                struct[field] = val.base_dict()
        return struct

    def upload(self, artifact_path: typing.Optional[str] = None):
        """
        internal, upload to target store
        :param artifact_path: required only for when generating target_path from artifact hash
        """
        src_path = self.spec.src_path
        body = self.get_body()
        if body:
            self._upload_body(body=body, artifact_path=artifact_path)
        else:
            if src_path and os.path.isfile(src_path):
                self._upload_file(source_path=src_path, artifact_path=artifact_path)

    def _upload_body(
        self, body, target=None, artifact_path: typing.Optional[str] = None
    ):
        body_hash = None
        if not target and not self.spec.target_path:
            if not mlrun.mlconf.artifacts.generate_target_path_from_artifact_hash:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    "Unable to resolve target path, no target path is defined and "
                    "mlrun.mlconf.artifacts.generate_target_path_from_artifact_hash is set to false"
                )
            body_hash, self.spec.target_path = self.resolve_body_target_hash_path(
                body, artifact_path
            )

        if mlrun.mlconf.artifacts.calculate_hash:
            self.metadata.hash = body_hash or calculate_blob_hash(body)
        self.spec.size = len(body)

        mlrun.datastore.store_manager.object(url=target or self.spec.target_path).put(
            body
        )

    def _upload_file(
        self,
        source_path: str,
        target_path: typing.Optional[str] = None,
        artifact_path: typing.Optional[str] = None,
    ):
        file_hash = None
        if not target_path and not self.spec.target_path:
            if not mlrun.mlconf.artifacts.generate_target_path_from_artifact_hash:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    "Unable to resolve target path, no target path is defined and "
                    "mlrun.mlconf.artifacts.generate_target_path_from_artifact_hash is set to false"
                )
            file_hash, self.spec.target_path = self.resolve_file_target_hash_path(
                source_path, artifact_path
            )
        if mlrun.mlconf.artifacts.calculate_hash:
            self.metadata.hash = file_hash or calculate_local_file_hash(source_path)
        self.spec.size = os.stat(source_path).st_size

        mlrun.datastore.store_manager.object(
            url=target_path or self.spec.target_path
        ).upload(source_path)

        if self._src_is_temp and os.path.exists(self.spec.src_path):
            # delete the temporary file if it was created for the upload
            os.remove(self.spec.src_path)

    def resolve_body_target_hash_path(
        self, body: typing.Union[bytes, str], artifact_path: str
    ) -> (str, str):
        """
        constructs the target path by calculating the artifact body hash
        :param body: artifact body to calculate hash on
        :param artifact_path: the base path for constructing the target path
        :return: [artifact_hash, target_path]
        """
        return self._resolve_target_hash_path(
            artifact_source=body,
            artifact_path=artifact_path,
            hash_method=calculate_blob_hash,
        )

    def resolve_file_target_hash_path(
        self, source_path: str, artifact_path: str
    ) -> (str, str):
        """
        constructs the target path by calculating the artifact source hash
        :param source_path: artifact file source path to calculate hash on
        :param artifact_path: the base path for constructing the target path
        :return: [artifact_hash, target_path]
        """
        return self._resolve_target_hash_path(
            artifact_source=source_path,
            artifact_path=artifact_path,
            hash_method=calculate_local_file_hash,
        )

    def _resolve_target_hash_path(
        self,
        artifact_source: typing.Union[bytes, str],
        artifact_path: str,
        hash_method: typing.Callable,
    ) -> (str, str):
        """
        constructs the target path by calculating the artifact source hash
        :param artifact_source: artifact to calculate hash on. May be path to source (str) or content (bytes)
        :param artifact_path: the base path for constructing the target path
        :param hash_method: the method which calculates the hash from the artifact source
        :return: [artifact_hash, target_path]
        """
        if not artifact_path:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "Unable to resolve file target hash path, artifact_path is not defined"
            )
        artifact_hash = hash_method(artifact_source)
        suffix = self._resolve_suffix()
        artifact_path = (
            artifact_path + "/" if not artifact_path.endswith("/") else artifact_path
        )
        target_path = f"{artifact_path}{artifact_hash}{suffix}"
        return artifact_hash, target_path

    def _resolve_suffix(self) -> str:
        suffix = "".join(pathlib.Path(self.spec.src_path or "").suffixes)
        if not suffix and self.spec.format:
            suffix = f".{self.spec.format}"
        return suffix

    # Following properties are for backwards compatibility with the ArtifactLegacy class. They should be
    # removed once we only work with the new Artifact structure.

    def is_inline(self):
        return self.spec._is_inline

    @property
    def inline(self):
        return self.spec.inline

    @inline.setter
    def inline(self, body):
        self.spec.inline = body

    @property
    def tag(self):
        return self.metadata.tag

    @tag.setter
    def tag(self, tag):
        self.metadata.tag = tag

    @property
    def key(self):
        return self.metadata.key

    @key.setter
    def key(self, key):
        self.metadata.key = key

    @property
    def src_path(self):
        return self.spec.src_path

    @src_path.setter
    def src_path(self, src_path):
        self.spec.src_path = src_path

    @property
    def target_path(self):
        return self.spec.target_path

    @target_path.setter
    def target_path(self, target_path):
        self.spec.target_path = target_path

    @property
    def producer(self):
        return self.spec.producer

    @producer.setter
    def producer(self, producer):
        self.spec.producer = producer

    @property
    def format(self):
        return self.spec.format

    @format.setter
    def format(self, format):
        self.spec.format = format

    @property
    def viewer(self):
        return self.spec.viewer

    @viewer.setter
    def viewer(self, viewer):
        self.spec.viewer = viewer

    @property
    def size(self):
        return self.spec.size

    @size.setter
    def size(self, size):
        self.spec.size = size

    @property
    def db_key(self):
        return self.spec.db_key

    @db_key.setter
    def db_key(self, db_key):
        self.spec.db_key = db_key

    @property
    def sources(self):
        return self.spec.sources

    @sources.setter
    def sources(self, sources):
        self.spec.sources = sources

    @property
    def extra_data(self):
        return self.spec.extra_data

    @extra_data.setter
    def extra_data(self, extra_data):
        self.spec.extra_data = extra_data

    @property
    def labels(self):
        return self.metadata.labels

    @labels.setter
    def labels(self, labels):
        self.metadata.labels = labels

    @property
    def iter(self):
        return self.metadata.iter

    @iter.setter
    def iter(self, iter):
        self.metadata.iter = iter

    @property
    def tree(self):
        return self.metadata.tree

    @tree.setter
    def tree(self, tree):
        self.metadata.tree = tree

    @property
    def project(self):
        return self.metadata.project

    @project.setter
    def project(self, project):
        self.metadata.project = project

    @property
    def hash(self):
        return self.metadata.hash

    @hash.setter
    def hash(self, hash):
        self.metadata.hash = hash

    @property
    def uid(self):
        return self.metadata.uid

    @uid.setter
    def uid(self, uid):
        self.metadata.uid = uid

    def generate_target_path(self, artifact_path, producer):
        return generate_target_path(self, artifact_path, producer)


class DirArtifactSpec(ArtifactSpec):
    _dict_fields = [
        "src_path",
        "target_path",
        "db_key",
        "producer",
    ]


class DirArtifact(Artifact):
    kind = "dir"

    @property
    def spec(self) -> DirArtifactSpec:
        return self._spec

    @spec.setter
    def spec(self, spec):
        self._spec = self._verify_dict(spec, "spec", DirArtifactSpec)

    @property
    def is_dir(self):
        return True

    def upload(self, artifact_path: typing.Optional[str] = None):
        """
        internal, upload to target store
        :param artifact_path: required only for when generating target_path from artifact hash
        """
        if not self.spec.src_path:
            raise mlrun.errors.MLRunInvalidArgumentError(
                "local/source path not specified"
            )

        files = os.listdir(self.spec.src_path)
        for file_name in files:
            file_path = os.path.join(self.spec.src_path, file_name)
            if not os.path.isfile(file_path):
                raise mlrun.errors.MLRunNotFoundError(
                    f"file {file_path} not found, cant upload"
                )

            if self.spec.target_path:
                target_path = os.path.join(self.spec.target_path, file_name)
            elif mlrun.mlconf.artifacts.generate_target_path_from_artifact_hash:
                _, target_path = self.resolve_file_target_hash_path(
                    source_path=file_path, artifact_path=artifact_path
                )
            else:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    "target path is not specified and mlrun.mlconf.artifacts.generate_target_path_from_artifact_hash "
                    "set to False"
                )

            mlrun.datastore.store_manager.object(url=target_path).upload(file_path)
            # add files of the directory to the extra data of the artifact with value of the target path
            self.spec.extra_data[file_name] = target_path


class LinkArtifactSpec(ArtifactSpec):
    _dict_fields = ArtifactSpec._dict_fields + [
        "link_iteration",
        "link_key",
        "link_tree",
    ]

    def __init__(
        self,
        src_path=None,
        target_path=None,
        link_iteration=None,
        link_key=None,
        link_tree=None,
    ):
        super().__init__(src_path, target_path)
        self.link_iteration = link_iteration
        self.link_key = link_key
        self.link_tree = link_tree


class LinkArtifact(Artifact):
    kind = "link"

    def __init__(
        self,
        key=None,
        target_path="",
        link_iteration=None,
        link_key=None,
        link_tree=None,
        project=None,
        metadata: ArtifactMetadata = None,
        spec: LinkArtifactSpec = None,
    ):
        super().__init__(
            key, target_path=target_path, project=project, metadata=metadata, spec=spec
        )
        self.spec.link_iteration = link_iteration
        self.spec.link_key = link_key
        self.spec.link_tree = link_tree

    @property
    def spec(self) -> LinkArtifactSpec:
        return self._spec

    @spec.setter
    def spec(self, spec):
        self._spec = self._verify_dict(spec, "spec", LinkArtifactSpec)


def calculate_blob_hash(data):
    if isinstance(data, str):
        data = data.encode()
    h = hashlib.sha1()
    h.update(data)
    return h.hexdigest()


def upload_extra_data(
    artifact: Artifact,
    extra_data: dict,
    prefix="",
    update_spec=False,
    artifact_path: typing.Optional[str] = None,
):
    """upload extra data to the artifact store"""
    if not extra_data:
        return
    target_path = artifact.target_path
    for key, item in extra_data.items():
        if item is ...:
            # Skip future links (packagers feature for linking artifacts before they are logged)
            continue

        if isinstance(item, bytes):
            if target_path:
                target = os.path.join(target_path, prefix + key)
            else:
                _, target = artifact.resolve_body_target_hash_path(
                    item, artifact_path=artifact_path
                )

            mlrun.datastore.store_manager.object(url=target).put(item)
            artifact.extra_data[prefix + key] = target
            continue

        if is_relative_path(item):
            src_path = (
                os.path.join(artifact.src_path, item) if artifact.src_path else item
            )
            if not os.path.isfile(src_path):
                raise ValueError(f"Extra data file {src_path} not found")

            if target_path:
                target = os.path.join(target_path, item)
            else:
                _, target = artifact.resolve_file_target_hash_path(
                    src_path, artifact_path=artifact_path
                )
            mlrun.datastore.store_manager.object(url=target).upload(src_path)
            artifact.extra_data[prefix + key] = target
            continue

        if update_spec:
            artifact.extra_data[prefix + key] = item


def get_artifact_meta(artifact):
    """return artifact object, and list of extra data items


    :param artifact:   artifact path (store://..) or DataItem

    :returns: artifact object, extra data dict

    """
    if hasattr(artifact, "artifact_url"):
        artifact = artifact.artifact_url

    if mlrun.datastore.is_store_uri(artifact):
        artifact_spec, _ = mlrun.datastore.store_manager.get_store_artifact(artifact)

    elif artifact.lower().endswith(".yaml"):
        data = mlrun.datastore.store_manager.object(url=artifact).get()
        spec = yaml.load(data, Loader=yaml.FullLoader)
        artifact_spec = mlrun.artifacts.dict_to_artifact(spec)

    else:
        raise ValueError(f"cant resolve artifact file for {artifact}")

    extra_dataitems = {}
    for k, v in artifact_spec.extra_data.items():
        extra_dataitems[k] = mlrun.datastore.store_manager.object(v, key=k)

    return artifact_spec, extra_dataitems


def generate_target_path(item: Artifact, artifact_path, producer):
    # path convention: artifact_path[/{run_name}]/{iter}/{key}.{suffix}
    # todo: add run_id here (vs in the .run() methods), support items dedup (by hash)
    artifact_path = artifact_path or ""
    if artifact_path and not artifact_path.endswith("/"):
        artifact_path += "/"
    if producer.kind == "run":
        artifact_path += f"{producer.name}/{item.iter or 0}/"

    suffix = "/"
    if not item.is_dir:
        # suffixes yields a list of suffixes, e.g. ['.tar', '.gz']
        # join them together to get the full suffix, e.g. '.tar.gz'
        suffix = "".join(pathlib.Path(item.src_path or "").suffixes)
        if not suffix and item.format:
            suffix = f".{item.format}"

    return f"{artifact_path}{item.key}{suffix}"


def fill_artifact_object_hash(object_dict, iteration=None, producer_id=None):
    # remove artifact related fields before calculating hash
    object_dict.setdefault("metadata", {})
    labels = object_dict["metadata"].pop("labels", None)
    object_updated_timestamp = object_dict["metadata"].pop("updated", None)

    artifact_cls = mlrun.artifacts.artifact_types.get(
        object_dict.get("kind", "artifact"), Artifact
    )()
    spec_fields_to_exclude = artifact_cls.spec._exclude_fields_from_uid_hash
    spec_fields_to_exclude_values = []
    object_dict.setdefault("spec", {})
    for field in spec_fields_to_exclude:
        spec_fields_to_exclude_values.append(object_dict["spec"].pop(field, None))

    # make sure we have a key, producer_id and iteration, as they determine the artifact uniqueness
    if not object_dict["metadata"].get("key"):
        raise ValueError("Artifact key is not set")
    object_dict["metadata"]["iter"] = iteration or object_dict["metadata"].get("iter")
    object_dict["metadata"]["tree"] = object_dict["metadata"].get("tree") or producer_id

    # calc hash and fill
    uid = mlrun.utils.helpers.fill_object_hash(object_dict, "uid")

    # restore original values
    if labels:
        object_dict["metadata"]["labels"] = labels
    if object_updated_timestamp:
        object_dict["metadata"]["updated"] = object_updated_timestamp
    for key, value in zip(spec_fields_to_exclude, spec_fields_to_exclude_values):
        if value is not None:
            object_dict["spec"][key] = value

    return uid


def verify_target_path(artifact: Artifact):
    if not artifact.get_target_path():
        raise mlrun.errors.MLRunInvalidArgumentError(
            f"artifact {artifact.uri} "
            f"does not have a valid/persistent offline target"
        )



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import abc
import os
from typing import Optional

from mlrun.config import config
from mlrun.runtimes.kubejob import KubejobRuntime
from mlrun.runtimes.pod import KubeResourceSpec
from mlrun.utils import update_in


class MPIResourceSpec(KubeResourceSpec):
    _dict_fields = KubeResourceSpec._dict_fields + ["mpi_args"]

    def __init__(
        self,
        command=None,
        args=None,
        image=None,
        mode=None,
        volumes=None,
        volume_mounts=None,
        env=None,
        resources=None,
        default_handler=None,
        entry_points=None,
        description=None,
        workdir=None,
        replicas=None,
        image_pull_policy=None,
        service_account=None,
        build=None,
        image_pull_secret=None,
        mpi_args=None,
        node_name=None,
        node_selector=None,
        affinity=None,
        priority_class_name=None,
        disable_auto_mount=False,
        pythonpath=None,
        tolerations=None,
        preemption_mode=None,
        security_context=None,
        state_thresholds=None,
        serving_spec=None,
        graph=None,
        parameters=None,
        track_models=None,
    ):
        super().__init__(
            command=command,
            image=image,
            mode=mode,
            build=build,
            entry_points=entry_points,
            description=description,
            workdir=workdir,
            default_handler=default_handler,
            volumes=volumes,
            volume_mounts=volume_mounts,
            env=env,
            resources=resources,
            replicas=replicas,
            image_pull_policy=image_pull_policy,
            service_account=service_account,
            image_pull_secret=image_pull_secret,
            args=args,
            node_name=node_name,
            node_selector=node_selector,
            affinity=affinity,
            priority_class_name=priority_class_name,
            disable_auto_mount=disable_auto_mount,
            pythonpath=pythonpath,
            tolerations=tolerations,
            preemption_mode=preemption_mode,
            security_context=security_context,
            state_thresholds=state_thresholds,
            serving_spec=serving_spec,
            graph=graph,
            parameters=parameters,
            track_models=track_models,
        )
        self.mpi_args = mpi_args or [
            "-x",
            "NCCL_SOCKET_NTHREADS=2",
            "-x",
            "NCCL_NSOCKS_PERTHREAD=8",
            "-x",
            "NCCL_MIN_NCHANNELS=4",
        ]


class AbstractMPIJobRuntime(KubejobRuntime, abc.ABC):
    kind = "mpijob"
    # nested i.e. hyper-param loop will use the same CRD/containers (vs CRD per iteration)
    _is_nested = True

    @property
    def spec(self) -> MPIResourceSpec:
        return self._spec

    @spec.setter
    def spec(self, spec):
        self._spec = self._verify_dict(spec, "spec", MPIResourceSpec)

    @staticmethod
    def _get_run_completion_updates(run: dict) -> dict:
        # TODO: add a 'workers' section in run objects state, each worker will update its state while
        #  the run state will be resolved by the server.
        # update the run object state if empty so that it won't default to 'created' state
        update_in(run, "status.state", "running", append=False, replace=False)
        return {}

    def with_tracing(
        self, log_file_path: Optional[str] = None, enable_cycle_markers: bool = False
    ):
        """Add Horovod Timeline activity tracking to the job to analyse
        its performance.

        The data will be saved as JSON to {log_file_path}. It can then be viewed via
        a trace viewer like chrome or edge's `edge://tracing`.

        More information can be found in the official documentation:
        https://horovod.readthedocs.io/en/latest/timeline_include.html

        Args:
            log_file_path (str, optional):         filepath for the json log file.
                                                   Defaults to <artifacts_path>/hvd_logs/trace.log.
            enable_cycle_markers (bool, optional): Add cycle markers to the log for
                                                   Tensor Fusion aid. Could make the trace very crowded.
                                                   Defaults to False.
        """

        log_path = (
            os.path.join(config.artifact_path, "hvd_logs", "trace.log")
            if log_file_path is None
            else log_file_path
        )
        horovod_timeline_settings = {
            "HOROVOD_TIMELINE": log_path,
            "HOROVOD_TIMELINE_MARK_CYCLES": int(enable_cycle_markers),
        }
        self.set_envs(horovod_timeline_settings)

    def with_autotune(
        self,
        log_file_path: Optional[str] = None,
        warmup_samples: Optional[int] = None,
        steps_per_sample: Optional[int] = None,
        bayes_opt_max_samples: Optional[int] = None,
        gaussian_process_noise: Optional[float] = None,
    ):
        """Adds an Autotuner to help optimize Horovod's Parameters for better performance.

        The autotuner will collect metrics and tune horovod's parameters while running using
        Bayesian optimiation. This may affect the performance of the run initially but after
        arriving to the best parameters should increase performance.

        Since autotuning imposes a tradeoff between early performance for better performance
        later on, It's advised to enable it when both:
        - Training should take a long timeout
        - Scaling efficiency was found lacking with the default settings

        More information can be found in the official documentation:
        https://horovod.readthedocs.io/en/latest/autotune_include.html

        Args:
            log_file_path (str, optional):            filepath for the csv log file.
                                                      Defaults to <artifacts_path>/hvd_logs/autotune.csv
            warmup_samples (int, optional):           number of discarded samples at the beginning of the training
                                                      process. Defaults to None.
            steps_per_sample (int, optional):         steps per sample. Defaults to None.
            bayes_opt_max_samples (int, optional):    maximum number of samples. Defaults to None.
            gaussian_process_noise (float, optional): Bayes optimizer's Alpha (noise regularization), to
                                                      account for network and resources variance.
                                                      Defaults to None.
        """

        log_path = (
            os.path.join(config.artifact_path, "hvd_logs", "autotune.csv")
            if log_file_path is None
            else log_file_path
        )
        horovod_autotune_settings = {
            "HOROVOD_AUTOTUNE": "1",
            "HOROVOD_AUTOTUNE_LOG": log_path,
        }
        if warmup_samples is not None:
            horovod_autotune_settings["autotune-warmup-samples"] = warmup_samples
        if steps_per_sample is not None:
            horovod_autotune_settings["autotune-steps-per-sample"] = steps_per_sample
        if bayes_opt_max_samples is not None:
            horovod_autotune_settings["autotune-bayes-opt-max-samples"] = (
                bayes_opt_max_samples
            )
        if gaussian_process_noise is not None:
            horovod_autotune_settings["autotune-gaussian-process-noise"] = (
                gaussian_process_noise
            )

        self.set_envs(horovod_autotune_settings)

    def set_mpi_args(self, args: list[str]) -> None:
        """Sets the runtime's mpi arguments to args.

        Parameters
        ----------
        args : typing.List[str]
            Arguments to be used for the mpi-operator

        Raises
        ------
        ValueError
            args is of type `List[str]` and can only accept `str` parameters.

        Example
        -------
        ```
        # Define the wanted MPI arguments
        mpi_args = []
        mpi_args.append("-x")
        mpi_args.append("NCCL_DEBUG=INFO")
        mpi_args.append("-x")
        mpi_args.append("NCCL_SOCKET_NTHREADS=2")
        mpi_args.append("-x")
        mpi_args.append("NCCL_NSOCKS_PERTHREAD=8")
        mpi_args.append("-x")
        mpi_args.append("NCCL_MIN_NCHANNELS=4")

        # Set the MPI arguments in the function
        fn.set_mpi_args(mpi_args)
        ```

        Notes
        -----
        * This will replace existing args.

        """

        # Verify that we are given only strings
        if not all([isinstance(arg, str) for arg in args]):
            raise ValueError(
                "Args is of type `List[str]` and can only accept `str` type params."
            )

        self.spec.mpi_args = args



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
from typing import Optional

from pydantic.v1 import BaseModel, Extra

import mlrun.common.types


class ObjectMetadata(BaseModel):
    name: str
    project: Optional[str]
    tag: Optional[str]
    labels: Optional[dict] = {}
    updated: Optional[datetime]
    created: Optional[datetime]
    uid: Optional[str]

    class Config:
        extra = Extra.allow


class ObjectStatus(BaseModel):
    state: Optional[str]

    class Config:
        extra = Extra.allow


class ObjectSpec(BaseModel):
    class Config:
        extra = Extra.allow


class LabelRecord(BaseModel):
    id: int
    name: str
    value: str

    class Config:
        orm_mode = True


class ObjectRecord(BaseModel):
    id: int
    name: str
    project: str
    uid: str
    updated: Optional[datetime] = None
    labels: list[LabelRecord]
    # state is extracted from the full status dict to enable queries
    state: Optional[str] = None
    full_object: Optional[dict] = None

    class Config:
        orm_mode = True


class ObjectKind(mlrun.common.types.StrEnum):
    project = "project"
    feature_set = "FeatureSet"
    background_task = "BackgroundTask"
    feature_vector = "FeatureVector"
    model_endpoint = "model-endpoint"
    hub_source = "HubSource"
    hub_item = "HubItem"
    hub_catalog = "HubCatalog"


class ObjectStatusState(mlrun.common.types.StrEnum):
    CREATED = "created"



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
from typing import Optional, Union

import mlrun
from mlrun.frameworks.sklearn.metric import Metric

from .._ml_common import MLArtifactsLibrary, MLPlan, PickleModelServer
from .metrics_library import MetricsLibrary
from .mlrun_interface import SKLearnMLRunInterface
from .model_handler import SKLearnModelHandler
from .utils import SKLearnTypes, SKLearnUtils

# Placeholders as the SciKit-Learn API is commonly used among all ML frameworks:
SKLearnModelServer = PickleModelServer
SKLearnArtifactsLibrary = MLArtifactsLibrary


def apply_mlrun(
    model: SKLearnTypes.ModelType = None,
    model_name: str = "model",
    tag: str = "",
    model_path: Optional[str] = None,
    modules_map: Optional[Union[dict[str, Union[None, str, list[str]]], str]] = None,
    custom_objects_map: Optional[Union[dict[str, Union[str, list[str]]], str]] = None,
    custom_objects_directory: Optional[str] = None,
    context: mlrun.MLClientCtx = None,
    artifacts: Optional[Union[list[MLPlan], list[str], dict[str, dict]]] = None,
    metrics: Optional[
        Union[
            list[Metric],
            list[SKLearnTypes.MetricEntryType],
            dict[str, SKLearnTypes.MetricEntryType],
        ]
    ] = None,
    x_test: SKLearnTypes.DatasetType = None,
    y_test: SKLearnTypes.DatasetType = None,
    sample_set: Union[SKLearnTypes.DatasetType, mlrun.DataItem, str] = None,
    y_columns: Optional[Union[list[str], list[int]]] = None,
    feature_vector: Optional[str] = None,
    feature_weights: Optional[list[float]] = None,
    labels: Optional[dict[str, Union[str, int, float]]] = None,
    parameters: Optional[dict[str, Union[str, int, float]]] = None,
    extra_data: Optional[dict[str, SKLearnTypes.ExtraDataType]] = None,
    auto_log: bool = True,
    **kwargs,
) -> SKLearnModelHandler:
    """
    Wrap the given model with MLRun's interface providing it with mlrun's additional features.

    :param model:                    The model to wrap. Can be loaded from the model path given as well.
    :param model_name:               The model name to use for storing the model artifact. Default: "model".
    :param tag:                      The model's tag to log with.
    :param model_path:               The model's store object path. Mandatory for evaluation (to know which model to
                                     update). If model is not provided, it will be loaded from this path.
    :param modules_map:              A dictionary of all the modules required for loading the model. Each key is a
                                     path to a module and its value is the object name to import from it. All the
                                     modules will be imported globally. If multiple objects needed to be imported
                                     from the same module a list can be given. The map can be passed as a path to a
                                     json file as well. For example:

                                     .. code-block:: python

                                         {
                                             "module1": None,  # import module1
                                             "module2": ["func1", "func2"],  # from module2 import func1, func2
                                             "module3.sub_module": "func3",  # from module3.sub_module import func3
                                         }

                                     If the model path given is of a store object, the modules map will be read from
                                     the logged modules map artifact of the model.
    :param custom_objects_map:       A dictionary of all the custom objects required for loading the model. Each key is
                                     a path to a python file and its value is the custom object name to import from it.
                                     If multiple objects needed to be imported from the same py file a list can be
                                     given. The map can be passed as a path to a json file as well. For example:

                                     .. code-block:: python

                                         {
                                             "/.../custom_model.py": "MyModel",
                                             "/.../custom_objects.py": ["object1", "object2"],
                                         }

                                     All the paths will be accessed from the given 'custom_objects_directory', meaning
                                     each py file will be read from 'custom_objects_directory/<MAP VALUE>'. If the model
                                     path given is of a store object, the custom objects map will be read from the
                                     logged custom object map artifact of the model.
                                     Notice: The custom objects will be imported in the order they came in this
                                     dictionary (or json). If a custom object is depended on another, make sure to
                                     put it below the one it relies on.
    :param custom_objects_directory: Path to the directory with all the python files required for the custom objects.
                                     Can be passed as a zip file as well (will be extracted during the run before
                                     loading the model). If the model path given is of a store object, the custom
                                     objects files will be read from the logged custom object artifact of the model.
    :param context:                  MLRun context to work with. If no context is given it will be retrieved via
                                     'mlrun.get_or_create_ctx(None)'
    :param artifacts:                A list of artifacts plans to produce during the run.
    :param metrics:                  A list of metrics to calculate during the run.
    :param x_test:                   The validation data for producing and calculating artifacts and metrics post
                                     training. Without this, validation will not be performed.
    :param y_test:                   The test data ground truth for producing and calculating artifacts and metrics post
                                     training or post predict / predict_proba.
    :param sample_set:               A sample set of inputs for the model for logging its stats along the model in
                                     favour of model monitoring. If not given the 'x_train' will be used by default.
    :param y_columns:                List of names of all the columns in the ground truth labels in case its a
                                     pd.DataFrame or a list of integers in case the dataset is a np.ndarray. If not
                                     given 'y_train' is given then the labels / indices in it will be used by default.
    :param feature_vector:           Feature store feature vector uri (store://feature-vectors/<project>/<name>[:tag])
    :param feature_weights:          List of feature weights, one per input column.
    :param labels:                   Labels to log with the model.
    :param parameters:               Parameters to log with the model.
    :param extra_data:               Extra data to log with the model.
    :param auto_log:                 Whether to apply MLRun's auto logging on the model. Auto logging will add the
                                     default artifacts and metrics to the lists of artifacts and metrics. Default:
                                     True.

    :return: The model handler initialized with the provided model.
    """
    # Get the default context:
    if context is None:
        context = mlrun.get_or_create_ctx(SKLearnMLRunInterface.DEFAULT_CONTEXT_NAME)

    # Create a model handler:
    model_handler_kwargs = (
        kwargs.pop("model_handler_kwargs") if "model_handler_kwargs" in kwargs else {}
    )
    handler = SKLearnModelHandler(
        model_name=model_name,
        model_path=model_path,
        model=model,
        context=context,
        modules_map=modules_map,
        custom_objects_map=custom_objects_map,
        custom_objects_directory=custom_objects_directory,
        **model_handler_kwargs,
    )

    # Load the model if it was not provided:
    if model is None:
        handler.load()
        model = handler.model

    # Set the handler's logging attributes:
    handler.set_tag(tag=tag)
    if sample_set is not None:
        handler.set_sample_set(sample_set=sample_set)
    if y_columns is not None:
        handler.set_target_columns(target_columns=y_columns)
    if feature_vector is not None:
        handler.set_feature_vector(feature_vector=feature_vector)
    if feature_weights is not None:
        handler.set_feature_weights(feature_weights=feature_weights)
    if labels is not None:
        handler.set_labels(to_add=labels)
    if parameters is not None:
        handler.set_parameters(to_add=parameters)
    if extra_data is not None:
        handler.set_extra_data(to_add=extra_data)

    # Add MLRun's interface to the model:
    SKLearnMLRunInterface.add_interface(obj=model)

    # Configure the logger:
    model.configure_logging(
        context=context,
        plans=SKLearnArtifactsLibrary.get_plans(
            artifacts=artifacts,
            context=context,
            include_default=auto_log,
            model=model,
            y=y_test,
        ),
        metrics=MetricsLibrary.get_metrics(
            metrics=metrics,
            context=context,
            include_default=auto_log,
            model=model,
            y=y_test,
        ),
        x_test=x_test,
        y_test=y_test,
        model_handler=handler,
    )

    return handler



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
from typing import Callable, Optional, Union

import torch
from torch import Tensor
from torch.nn import Module, Parameter
from torch.optim import Optimizer
from torch.utils.data import DataLoader
from torch.utils.tensorboard import SummaryWriter
from torch.utils.tensorboard.summary import hparams

import mlrun

from ..._dl_common.loggers import TensorboardLogger
from ..utils import PyTorchTypes
from .logging_callback import LoggingCallback


class _MLRunSummaryWriter(SummaryWriter):
    """
    A slightly edited torch's SummaryWriter class to overcome the hyperparameter logging problem (creating a new event
    per call).
    """

    def add_hparams(
        self, hparam_dict, metric_dict, hparam_domain_discrete=None, run_name=None
    ):
        """
        Log the given hyperparameters to the same event file that is currently open.

        :param hparam_dict:            The static hyperparameters to simply log to the 'hparams' table.
        :param metric_dict:            The metrics and dynamic hyper parameters to link with the plots.
        :param hparam_domain_discrete: Not used in this SummaryWriter.
        :param run_name:               Not used in this SummaryWriter.
        """
        torch._C._log_api_usage_once("tensorboard.logging.add_hparams")
        if not isinstance(hparam_dict, dict) or not isinstance(metric_dict, dict):
            raise TypeError("hparam_dict and metric_dict should be dictionary.")
        exp, ssi, sei = hparams(hparam_dict, metric_dict)
        self._get_file_writer().add_summary(exp)
        self._get_file_writer().add_summary(ssi)
        self._get_file_writer().add_summary(sei)


class _PyTorchTensorboardLogger(TensorboardLogger):
    """
    The PyTorch framework implementation of the 'TensorboardLogger'.
    """

    def __init__(
        self,
        statistics_functions: list[
            Callable[[Union[Parameter]], Union[float, Parameter]]
        ],
        context: mlrun.MLClientCtx = None,
        tensorboard_directory: Optional[str] = None,
        run_name: Optional[str] = None,
        update_frequency: Union[int, str] = "epoch",
    ):
        """
        Initialize a tensorboard logger callback with the given configuration. At least one of 'context' and
        'tensorboard_directory' must be given.

        :param statistics_functions:  A list of statistics functions to calculate at the end of each epoch on the
                                      tracked weights. Only relevant if weights are being tracked. The functions in
                                      the list must accept one Weight and return a float (or float convertible) value.
        :param context:               A MLRun context to use for logging into the user's tensorboard directory. The
                                      context parameters can be logged as static hyperparameters as well.
        :param tensorboard_directory: If context is not given, or if wished to set the directory even with context,
                                      this will be the output for the event logs of tensorboard. If not given, context
                                      must be provided as the default tensorboard output directory will be:
                                      /User/.tensorboard/<PROJECT_NAME> or if working on local, the set artifacts path.
        :param run_name:              This experiment run name. Each run name will be indexed at the end of the name so
                                      each experiment will be numbered automatically. If a context was given, the
                                      context's uid will be added instead of an index. If a run name was not given the
                                      current time stamp will be used.
        :param update_frequency:      Per how many iterations (batches) the callback should write the tracked values to
                                      tensorboard. Can be passed as a string equal to 'epoch' for per epoch and 'batch'
                                      for per single batch, or as an integer specifying per how many iterations to
                                      update. Notice that writing to tensorboard too frequently may cause the training
                                      to be slower. Default: 'epoch'.
        """
        super().__init__(
            statistics_functions=statistics_functions,
            context=context,
            tensorboard_directory=tensorboard_directory,
            run_name=run_name,
            update_frequency=update_frequency,
        )

        # Setup the tensorboard writer property:
        self._summary_writer = None

    def open(self):
        """
        Create the output path and initialize the tensorboard file writer.
        """
        # Create the output path:
        self._create_output_path()

        # Use the output path to initialize the tensorboard writer:
        self._summary_writer = _MLRunSummaryWriter(log_dir=self._output_path)

    def write_model_to_tensorboard(self, model: Module, input_sample: Tensor):
        """
        Write the given model as a graph in tensorboard.

        :param model:        The model to write.
        :param input_sample: An input sample for writing the model.
        """
        self._summary_writer.add_graph(
            model=model,
            input_to_model=input_sample,
        )

    def write_parameters_table_to_tensorboard(self):
        """
        Write the summaries, static and dynamic hyperparameters to the table in tensorboard's hparams section. This
        method is called once for creating the hparams table.
        """
        # Check if needed to track hyperparameters:
        if (
            len(self._static_hyperparameters) == 0
            and len(self._dynamic_hyperparameters) == 0
        ):
            return

        # Prepare the hyperparameters values:
        non_graph_parameters = {"Date": str(datetime.now()).split(".")[0]}
        for parameter, value in self._static_hyperparameters.items():
            non_graph_parameters[parameter] = value

        # Prepare the summaries values and the dynamic hyperparameters values:
        graph_parameters = {}
        for metric in self._training_summaries:
            graph_parameters[f"{self._Sections.SUMMARY}/training_{metric}"] = 0.0
        for metric in self._validation_summaries:
            graph_parameters[f"{self._Sections.SUMMARY}/validation_{metric}"] = 0.0
        for parameter, epochs in self._dynamic_hyperparameters.items():
            graph_parameters[f"{self._Sections.HYPERPARAMETERS}/{parameter}"] = epochs[
                -1
            ]

        # Write the hyperparameters and summaries table:
        self._summary_writer.add_hparams(non_graph_parameters, graph_parameters)

    def flush(self):
        """
        Make sure all values were written to the directory logs so it will be available live.
        """
        self._summary_writer.flush()

    def _write_text_to_tensorboard(self, tag: str, text: str, step: int):
        """
        Write text to tensorboard's text section. Summary information of this training / validation run will be logged
        to tensorboard using this method.

        :param tag:  The tag of the text (box it will be appearing under).
        :param text: The text to write.
        :param step: The iteration / epoch the text belongs to.
        """
        self._summary_writer.add_text(
            tag=tag,
            text_string=text,
            global_step=step,
        )

    def _write_scalar_to_tensorboard(self, name: str, value: float, step: int):
        """
        Write the scalar's value into its plot.

        :param name:  The plot's name.
        :param value: The value to add to the plot.
        :param step:  The iteration / epoch the value belongs to.
        """
        self._summary_writer.add_scalar(
            tag=name,
            scalar_value=value,
            global_step=step,
        )

    def _write_weight_histogram_to_tensorboard(
        self, name: str, weight: Parameter, step: int
    ):
        """
        Write the current state of the weights as histograms to tensorboard.

        :param name:   The weight's name.
        :param weight: The weight to write its histogram.
        :param step:   The iteration / epoch the weight's histogram state belongs to.
        """
        self._summary_writer.add_histogram(
            tag=name,
            values=weight,
            global_step=step,
        )

    def _write_weight_image_to_tensorboard(
        self, name: str, weight: Parameter, step: int
    ):
        """
        Log the current state of the weights as images to tensorboard.

        :param name:   The weight's name.
        :param weight: The weight to write its image.
        :param step:   The iteration / epoch the weight's image state belongs to.
        """
        raise NotImplementedError


class TensorboardLoggingCallback(LoggingCallback):
    """
    Callback for logging data during training / evaluation to tensorboard. the available data in tensorboard will be:

    * Summary text of the run with a hyperlink to the MLRun log if it was done.
    * Hyperparameters tuning table: static hyperparameters, dynamic hyperparameters and epoch validation summaries.
    * Plots:

      * Per iteration (batch) plot for the training and validation metrics.
      * Per epoch plot for the dynamic hyperparameters and validation summaries results.
      * Per epoch weights statistics for each weight and statistic.

    * Histograms per epoch for each of the logged weights.
    * Distributions per epoch for each of the logged weights.
    * Images per epoch for each of the logged weights.
    * Model architecture graph.

    All the collected data will be available in this callback post the training / validation process and can be accessed
    via the 'training_results', 'validation_results', 'static_hyperparameters', 'dynamic_hyperparameters', 'summaries',
    'weights', 'weights_mean' and 'weights_std' properties.
    """

    def __init__(
        self,
        context: mlrun.MLClientCtx = None,
        tensorboard_directory: Optional[str] = None,
        run_name: Optional[str] = None,
        weights: Union[bool, list[str]] = False,
        statistics_functions: Optional[
            list[Callable[[Union[Parameter, Tensor]], Union[float, Tensor]]]
        ] = None,
        dynamic_hyperparameters: Optional[
            dict[
                str,
                tuple[
                    str,
                    Union[
                        list[Union[str, int]], Callable[[], PyTorchTypes.TrackableType]
                    ],
                ],
            ]
        ] = None,
        static_hyperparameters: Optional[
            dict[
                str,
                Union[PyTorchTypes.TrackableType, tuple[str, list[Union[str, int]]]],
            ]
        ] = None,
        update_frequency: Union[int, str] = "epoch",
        auto_log: bool = False,
    ):
        """
        Initialize a tensorboard logging callback with the given weights, hyperparameters and logging configurations.
        Note that at least one of 'context' and 'tensorboard_directory' must be given.

        :param context:                 A mlrun context to use for logging into the user's tensorboard directory.
        :param tensorboard_directory:   If context is not given, or if wished to set the directory even with context,
                                        this will be the output for the event logs of tensorboard.
        :param run_name:                This experiment run name. Each run name will be indexed at the end of the name
                                        so each experiment will be numbered automatically. If a context was given, the
                                        context's uid will be added instead of an index. If a run name was not given the
                                        current time in the following format: 'YYYY-mm-dd_HH:MM:SS'.
        :param weights:                 If wished to track weights to draw their histograms and calculate statistics per
                                        epoch, the weights names should be passed here. Note that each name given will
                                        be searched as 'if <NAME> in <WEIGHT_NAME>' so a simple module name will be
                                        enough to catch his weights. A boolean value can be passed to track all weights.
                                        Default: False.
        :param statistics_functions:    A list of statistics functions to calculate at the end of each epoch on the
                                        tracked weights. Only relevant if weights are being tracked. The functions in
                                        the list must accept one Parameter (or Tensor) and return a float (or float
                                        convertible) value. The default statistics are 'mean' and 'std'. To get the
                                        default functions list for appending additional functions you can access it via
                                        'TensorboardLoggingCallback.get_default_weight_statistics_list()'. To not track
                                        statistics at all simply pass an empty list '[]'.
        :param dynamic_hyperparameters: If needed to track a hyperparameter dynamically (sample it each epoch) it should
                                        be passed here. The parameter expects a dictionary where the keys are the
                                        hyperparameter chosen names and the values are tuples of object key and a list
                                        with the key chain. A key chain is a list of keys and indices to know how to
                                        access the needed hyperparameter. If the hyperparameter is not of accessible
                                        from any of the HyperparametersKeys, a custom callable method can be passed in
                                        the tuple instead of the key chain when providing the word
                                        HyperparametersKeys.CUSTOM. For example, to track the 'lr' attribute of
                                        an optimizer and a custom parameter, one should pass:
                                        {
                                            "lr": (HyperparametersKeys.OPTIMIZER, ["param_groups", 0, "lr"]),
                                            "custom parameter": (HyperparametersKeys.CUSTOM, get_custom_parameter)
                                        }
        :param static_hyperparameters:  If needed to track a hyperparameter one time per run it should be passed here.
                                        The parameter expects a dictionary where the keys are the
                                        hyperparameter chosen names and the values are the hyperparameter static value
                                        or a tuple of object key and a list with the key chain just like the dynamic
                                        hyperparameter. For example, to track the 'epochs' of an experiment run, one
                                        should pass:
                                        {
                                            "epochs": 7
                                        }
        :param update_frequency:        Per how many iterations (batches) the callback should write the tracked values
                                        to tensorboard. Can be passed as a string equal to 'epoch' for per epoch and
                                        'batch' for per single batch, or as an integer specifying per how many
                                        iterations to update. Notice that writing to tensorboard too frequently may
                                        cause the training to be slower. Default: 'epoch'.
        :param auto_log:                Whether or not to enable auto logging, trying to track common static and dynamic
                                        hyperparameters.

        :raise MLRunInvalidArgumentError: In case both 'context' and 'tensorboard_directory' parameters were not given
                                          or the 'update_frequency' was incorrect.
        """
        super().__init__(
            dynamic_hyperparameters=dynamic_hyperparameters,
            static_hyperparameters=static_hyperparameters,
            auto_log=auto_log,
        )

        # Replace the logger with an MLRunLogger:
        del self._logger
        self._logger = _PyTorchTensorboardLogger(
            statistics_functions=(
                statistics_functions
                if statistics_functions is not None
                else self.get_default_weight_statistics_list()
            ),
            context=context,
            tensorboard_directory=tensorboard_directory,
            run_name=run_name,
            update_frequency=update_frequency,
        )

        # Save the configurations:
        self._tracked_weights = weights

    def get_weights(self) -> dict[str, Parameter]:
        """
        Get the weights tensors tracked. The weights will be stored in a dictionary where each key is the weight's name
        and the value is the weight's parameter (tensor).

        :return: The weights.
        """
        return self._logger.weights

    def get_weights_statistics(self) -> dict[str, dict[str, list[float]]]:
        """
        Get the weights mean results logged. The results will be stored in a dictionary where each key is the weight's
        name and the value is a list of mean values per epoch.

        :return: The weights mean results.
        """
        return self._logger.weight_statistics

    @staticmethod
    def get_default_weight_statistics_list() -> (
        list[Callable[[Union[Parameter, Tensor]], Union[float, Tensor]]]
    ):
        """
        Get the default list of statistics functions being applied on the tracked weights each epoch.

        :return: The default statistics functions list.
        """
        return [torch.mean, torch.std]

    def on_setup(
        self,
        model: Module = None,
        training_set: DataLoader = None,
        validation_set: DataLoader = None,
        loss_function: Module = None,
        optimizer: Optimizer = None,
        metric_functions: Optional[list[PyTorchTypes.MetricFunctionType]] = None,
        scheduler=None,
    ):
        """
        Storing all the given objects in the callback's objects dictionary and load the weights parameters as given in
        the callback's initialization.

        :param model:            The model to be stored in this callback.
        :param training_set:     The training set to be stored in this callback.
        :param validation_set:   The validation set to be stored in this callback.
        :param loss_function:    The loss function to be stored in this callback.
        :param optimizer:        The optimizer to be stored in this callback.
        :param metric_functions: The metric functions to be stored in this callback.
        :param scheduler:        The scheduler to be stored in this callback.
        """
        super().on_setup(
            model=model,
            training_set=training_set,
            validation_set=validation_set,
            loss_function=loss_function,
            optimizer=optimizer,
            metric_functions=metric_functions,
            scheduler=scheduler,
        )

        # Start the tensorboard logger:
        self._logger.open()

        # Collect the weights for drawing histograms according to the stored configuration:
        if self._tracked_weights is False:
            return

        # Log the weights:
        for weight_name, weight_parameter in self._objects[
            self._ObjectKeys.MODEL
        ].named_parameters():
            collect = False
            if self._tracked_weights is True:  # Collect all weights
                collect = True
            else:
                for tag in self._tracked_weights:  # Collect by given name
                    if tag in weight_name:
                        collect = True
                        break
            if collect:
                self._logger.log_weight(
                    weight_name=weight_name, weight_holder=weight_parameter
                )

        # Log the weights statistics:
        self._logger.log_weights_statistics()

    def on_run_begin(self):
        """
        After the run begins, this method will be called to setup the weights, results and hyperparameters dictionaries
        for logging. Epoch 0 (pre-run state) will be logged here.
        """
        # Setup all the results and hyperparameters dictionaries:
        super().on_run_begin()

        # Log the initial summary of the run:
        self._logger.write_initial_summary_text()

        # Log the model:
        self._logger.write_model_to_tensorboard(
            model=self._objects[self._ObjectKeys.MODEL],
            input_sample=next(self._objects[self._ObjectKeys.TRAINING_SET].__iter__())[
                0
            ],
        )

        # Write the hyperparameters table:
        self._logger.write_parameters_table_to_tensorboard()

        # Write the initial dynamic hyperparameters values:
        self._logger.write_dynamic_hyperparameters()

        # Write the initial weights data:
        self._logger.write_weights_statistics()
        self._logger.write_weights_histograms()
        self._logger.write_weights_images()

    def on_run_end(self):
        """
        Before the run ends, this method will be called to log the context summary.
        """
        # Write the final summary of the run:
        self._logger.write_final_summary_text()

        super().on_run_end()

    def on_validation_end(
        self, loss_value: PyTorchTypes.MetricValueType, metric_values: list[float]
    ):
        """
        Before the validation (in a training case it will be per epoch) ends, this method will be called to log the
        validation results summaries.

        :param loss_value:    The loss summary of this validation.
        :param metric_values: The metrics summaries of this validation.
        """
        super().on_validation_end(loss_value=loss_value, metric_values=metric_values)

        # Check if this run was part of an evaluation:
        if not self._is_training:
            # Write the remaining epoch iterations results:
            self._logger.write_validation_results(ignore_update_frequency=True)
            # Write the epoch loss and metrics summaries to their graphs:
            self._logger.write_validation_summaries()
            # Make sure all values were written to the directory logs:
            self._logger.flush()

    def on_epoch_end(self, epoch: int):
        """
        Before the given epoch ends, this method will be called to log the dynamic hyperparameters as needed. All of the
        per epoch plots (loss and metrics summaries, dynamic hyperparameters, weights histograms and statistics) will
        log this epoch's tracked values.

        :param epoch: The epoch that has just ended.
        """
        super().on_epoch_end(epoch=epoch)

        # Log the weights statistics:
        self._logger.log_weights_statistics()

        # Write the remaining epoch iterations results:
        self._logger.write_training_results(ignore_update_frequency=True)
        self._logger.write_validation_results(ignore_update_frequency=True)

        # Write the epoch text summary:
        self._logger.write_epoch_summary_text()

        # Write the epoch loss and metrics summaries to their graphs:
        self._logger.write_training_summaries()
        self._logger.write_validation_summaries()

        # Write the epoch dynamic hyperparameters values to their graphs:
        self._logger.write_dynamic_hyperparameters()

        # Write the weight histograms, images and statistics for all the tracked weights:
        self._logger.write_weights_statistics()
        self._logger.write_weights_histograms()
        self._logger.write_weights_images()

        # Make sure all values were written to the directory logs:
        self._logger.flush()

    def on_train_batch_end(self, batch: int, x: Tensor, y_true: Tensor, y_pred: Tensor):
        """
        Before the training of the given batch ends, this method will be called to log the batch's loss and metrics
        results to their per iteration plots.

        :param batch:  The current batch iteration of when this method is called.
        :param x:      The input part of the current batch.
        :param y_true: The true value part of the current batch.
        :param y_pred: The prediction (output) of the model for this batch's input ('x').
        """
        super().on_train_batch_end(batch=batch, x=x, y_true=y_true, y_pred=y_pred)

        # Write the batch loss and metrics results to their graphs:
        self._logger.write_training_results()

    def on_validation_batch_end(
        self, batch: int, x: Tensor, y_true: Tensor, y_pred: Tensor
    ):
        """
        Before the validation of the given batch ends, this method will be called to log the batch's loss and metrics
        results to their per iteration plots.

        :param batch:  The current batch iteration of when this method is called.
        :param x:      The input part of the current batch.
        :param y_true: The true value part of the current batch.
        :param y_pred: The prediction (output) of the model for this batch's input ('x').
        """
        super().on_validation_batch_end(batch=batch, x=x, y_true=y_true, y_pred=y_pred)

        # Write the batch loss and metrics results to their graphs:
        self._logger.write_validation_results()



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mlrun
from mlrun.data_types.to_pandas import spark_df_to_pandas
from mlrun.datastore.sources import ParquetSource
from mlrun.datastore.targets import get_offline_target
from mlrun.runtimes import RemoteSparkRuntime
from mlrun.runtimes.sparkjob import Spark3Runtime
from mlrun.utils.helpers import additional_filters_warning

from .base import BaseMerger


class SparkFeatureMerger(BaseMerger):
    engine = "spark"
    support_offline = True

    def __init__(self, vector, **engine_args):
        super().__init__(vector, **engine_args)
        self.spark = engine_args.get("spark", None)
        self.named_view = engine_args.get("named_view", False)
        self._pandas_df = None

    def to_spark_df(self, session, path):
        return session.read.load(path)

    def _unpersist_df(self, df):
        df.unpersist()

    def _asof_join(
        self,
        entity_df,
        entity_timestamp_column: str,
        featureset_name: str,
        featureset_timstamp: str,
        featureset_df: list,
        left_keys: list,
        right_keys: list,
    ):
        """Perform an as of join between entity and featureset.
        Join conditions:
        Args:
            entity_df (DataFrame): Spark dataframe representing the entities, to be joined with
                the feature tables.
            entity_timestamp_column (str): Column name in entity_df which represents
                event timestamp.
            featureset (Dataframe): Spark dataframe representing the feature table.
            featureset (FeatureSet): Feature set specification, which provides information on
                how the join should be performed, such as the entity primary keys.
        Returns:
            DataFrame: Join result, which contains all the original columns from entity_df, as well
                as all the features specified in featureset, where the feature columns will
                be prefixed with featureset_df name.
                :param featureset_name:
                :param featureset_timstamp:
        """

        from pyspark.sql import Window
        from pyspark.sql.functions import col, monotonically_increasing_id, row_number

        entity_with_id = entity_df.withColumn("_row_nr", monotonically_increasing_id())
        rename_right_keys = {}
        for key in right_keys + [featureset_timstamp]:
            if key in entity_df.columns:
                rename_right_keys[key] = f"ft__{key}"
        # get columns for projection
        projection = [
            col(col_name).alias(rename_right_keys.get(col_name, col_name))
            for col_name in featureset_df.columns
        ]

        aliased_featureset_df = featureset_df.select(projection)
        right_timestamp = rename_right_keys.get(
            featureset_timstamp, featureset_timstamp
        )

        # set join conditions
        join_cond = (
            entity_with_id[entity_timestamp_column]
            >= aliased_featureset_df[right_timestamp]
        )

        # join based on entities
        for key_l, key_r in zip(left_keys, right_keys):
            join_cond = join_cond & (
                entity_with_id[key_l]
                == aliased_featureset_df[rename_right_keys.get(key_r, key_r)]
            )

        conditional_join = entity_with_id.join(
            aliased_featureset_df, join_cond, "leftOuter"
        )

        window = Window.partitionBy("_row_nr").orderBy(
            col(right_timestamp).desc(),
        )
        filter_most_recent_feature_timestamp = conditional_join.withColumn(
            "_rank", row_number().over(window)
        ).filter(col("_rank") == 1)

        for key in right_keys + [featureset_timstamp]:
            if key in entity_df.columns + [entity_timestamp_column]:
                filter_most_recent_feature_timestamp = (
                    filter_most_recent_feature_timestamp.drop(
                        aliased_featureset_df[f"ft__{key}"]
                    )
                )
        return filter_most_recent_feature_timestamp.drop("_row_nr", "_rank").orderBy(
            col(entity_timestamp_column)
        )

    def _join(
        self,
        entity_df,
        entity_timestamp_column: str,
        featureset_name,
        featureset_timestamp,
        featureset_df,
        left_keys: list,
        right_keys: list,
    ):
        """
        spark dataframes join

        :param entity_df (DataFrame): Spark dataframe representing the entities, to be joined with
            the feature tables.
        :param entity_timestamp_column (str): Column name in entity_df which represents
            event timestamp.
        :param featureset_df (Dataframe): Spark dataframe representing the feature table.
        :param featureset_name:
        :param featureset_timestamp:

        Returns:
            DataFrame: Join result, which contains all the original columns from entity_df, as well
                as all the features specified in featureset, where the feature columns will
                be prefixed with featureset_df name.
        """
        if left_keys != right_keys:
            join_cond = [
                entity_df[key_l] == featureset_df[key_r]
                for key_l, key_r in zip(left_keys, right_keys)
            ]
        else:
            join_cond = left_keys

        merged_df = entity_df.join(
            featureset_df,
            join_cond,
            how=self._join_type,
        )
        return merged_df

    def get_df(self, to_pandas=True):
        if to_pandas:
            if self._pandas_df is None:
                df = spark_df_to_pandas(self._result_df)
                self._pandas_df = df
                self._set_indexes(self._pandas_df)
            return self._pandas_df

        return self._result_df

    @classmethod
    def get_default_image(cls, kind):
        if kind == Spark3Runtime.kind:
            return Spark3Runtime._get_default_deployed_mlrun_image_name(with_gpu=False)
        elif kind == RemoteSparkRuntime.kind:
            return RemoteSparkRuntime.default_image
        else:
            raise mlrun.errors.MLRunInvalidArgumentError(f"Unsupported kind '{kind}'")

    def _create_engine_env(self):
        from pyspark.sql import SparkSession

        if self.spark is None:
            # create spark context
            self.spark = (
                SparkSession.builder.appName(
                    f"vector-merger-{self.vector.metadata.name}"
                )
                .config("spark.driver.memory", "2g")
                .getOrCreate()
            )

    def _get_engine_df(
        self,
        feature_set,
        feature_set_name,
        column_names=None,
        start_time=None,
        end_time=None,
        time_column=None,
        additional_filters=None,
    ):
        source_kwargs = {}
        if feature_set.spec.passthrough:
            if not feature_set.spec.source:
                raise mlrun.errors.MLRunNotFoundError(
                    f"passthrough feature set {feature_set_name} with no source"
                )
            source_kind = feature_set.spec.source.kind
            source_path = feature_set.spec.source.path
            source_kwargs.update(feature_set.spec.source.attributes)
            source_kwargs.pop("additional_filters", None)
        else:
            target = get_offline_target(feature_set)
            if not target:
                raise mlrun.errors.MLRunInvalidArgumentError(
                    f"feature set {feature_set_name} does not have offline targets"
                )
            source_kind = target.kind
            source_path = target.get_target_path()
            source_kwargs = target.source_spark_attributes
        # handling case where there are multiple feature sets and user creates vector where
        # entity_timestamp_column is from a specific feature set (can't be entity timestamp)
        source_driver = mlrun.datastore.sources.source_kind_to_driver[source_kind]

        if source_driver != ParquetSource:
            additional_filters_warning(additional_filters, source_driver)
            additional_filters = None
        additional_filters_dict = (
            {"additional_filters": additional_filters} if additional_filters else {}
        )
        source = source_driver(
            name=self.vector.metadata.name,
            path=source_path,
            time_field=time_column,
            start_time=start_time,
            end_time=end_time,
            **additional_filters_dict,
            **source_kwargs,
        )

        columns = column_names + [ent.name for ent in feature_set.spec.entities]
        if (
            feature_set.spec.timestamp_key
            and feature_set.spec.timestamp_key not in columns
        ):
            columns.append(feature_set.spec.timestamp_key)

        return source.to_spark_df(
            self.spark,
            named_view=self.named_view,
            time_field=time_column,
            columns=columns,
        )

    def _rename_columns_and_select(
        self,
        df,
        rename_col_dict,
        columns=None,
    ):
        from pyspark.sql.functions import col

        return df.select(
            [
                col(name).alias(rename_col_dict.get(name, name))
                for name in columns or rename_col_dict.keys()
            ]
        )

    def _drop_columns_from_result(self):
        self._result_df = self._result_df.drop(*self._drop_columns)

    def _filter(self, query):
        self._result_df = self._result_df.filter(query)

    def _order_by(self, order_by_active):
        from pyspark.sql.functions import col

        self._result_df = self._result_df.orderBy(
            *[col(col_name).asc_nulls_last() for col_name in order_by_active]
        )

    def _convert_entity_rows_to_engine_df(self, entity_rows):
        if entity_rows is not None and not hasattr(entity_rows, "rdd"):
            return self.spark.createDataFrame(entity_rows)
        return entity_rows



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

import pydantic.v1

from .auth import AuthorizationResourceTypes, Credentials
from .object import (
    LabelRecord,
    ObjectKind,
    ObjectMetadata,
    ObjectRecord,
    ObjectSpec,
    ObjectStatus,
)


class FeatureStoreBaseModel(pydantic.v1.BaseModel):
    """
    Intermediate base class, in order to override pydantic's configuration, as per
    https://docs.pydantic.dev/1.10/usage/model_config/#change-behaviour-globally
    """

    class Config:
        copy_on_model_validation = "none"


class Feature(FeatureStoreBaseModel):
    name: str
    value_type: str
    labels: Optional[dict] = {}

    class Config:
        extra = pydantic.v1.Extra.allow


class Entity(FeatureStoreBaseModel):
    name: str
    value_type: str
    labels: Optional[dict] = {}

    class Config:
        extra = pydantic.v1.Extra.allow


class FeatureSetSpec(ObjectSpec):
    entities: list[Entity] = []
    features: list[Feature] = []
    engine: Optional[str] = pydantic.v1.Field(default="storey")


class FeatureSet(FeatureStoreBaseModel):
    kind: ObjectKind = pydantic.v1.Field(ObjectKind.feature_set, const=True)
    metadata: ObjectMetadata
    spec: FeatureSetSpec
    status: ObjectStatus

    @staticmethod
    def get_authorization_resource_type():
        return AuthorizationResourceTypes.feature_set


class EntityRecord(FeatureStoreBaseModel):
    name: str
    value_type: str
    labels: list[LabelRecord]

    class Config:
        orm_mode = True


class FeatureRecord(FeatureStoreBaseModel):
    name: str
    value_type: str
    labels: list[LabelRecord]

    class Config:
        orm_mode = True


class FeatureSetRecord(ObjectRecord):
    entities: list[EntityRecord]
    features: list[FeatureRecord]

    class Config:
        orm_mode = True


class FeatureSetsOutput(FeatureStoreBaseModel):
    feature_sets: list[FeatureSet]


class FeatureSetsTagsOutput(FeatureStoreBaseModel):
    tags: list[str] = []


class FeatureSetDigestSpec(FeatureStoreBaseModel):
    entities: list[Entity]
    features: list[Feature]


class FeatureSetDigestOutput(FeatureStoreBaseModel):
    metadata: ObjectMetadata
    spec: FeatureSetDigestSpec


class FeatureSetDigestSpecV2(FeatureStoreBaseModel):
    entities: list[Entity]


class FeatureSetDigestOutputV2(FeatureStoreBaseModel):
    feature_set_index: int
    metadata: ObjectMetadata
    spec: FeatureSetDigestSpecV2


class FeatureListOutput(FeatureStoreBaseModel):
    feature: Feature
    feature_set_digest: FeatureSetDigestOutput


class FeaturesOutput(FeatureStoreBaseModel):
    features: list[FeatureListOutput]


class FeaturesOutputV2(FeatureStoreBaseModel):
    features: list[Feature]
    feature_set_digests: list[FeatureSetDigestOutputV2]


class EntityListOutput(FeatureStoreBaseModel):
    entity: Entity
    feature_set_digest: FeatureSetDigestOutput


class EntitiesOutputV2(FeatureStoreBaseModel):
    entities: list[Entity]
    feature_set_digests: list[FeatureSetDigestOutputV2]


class EntitiesOutput(FeatureStoreBaseModel):
    entities: list[EntityListOutput]


class FeatureVector(FeatureStoreBaseModel):
    kind: ObjectKind = pydantic.v1.Field(ObjectKind.feature_vector, const=True)
    metadata: ObjectMetadata
    spec: ObjectSpec
    status: ObjectStatus

    @staticmethod
    def get_authorization_resource_type():
        return AuthorizationResourceTypes.feature_vector


class FeatureVectorRecord(ObjectRecord):
    pass


class FeatureVectorsOutput(FeatureStoreBaseModel):
    feature_vectors: list[FeatureVector]


class FeatureVectorsTagsOutput(FeatureStoreBaseModel):
    tags: list[str] = []


class DataSource(FeatureStoreBaseModel):
    kind: str
    name: str
    path: str

    class Config:
        extra = pydantic.v1.Extra.allow


class DataTarget(FeatureStoreBaseModel):
    kind: str
    name: str
    path: Optional[str]

    class Config:
        extra = pydantic.v1.Extra.allow


class FeatureSetIngestInput(FeatureStoreBaseModel):
    source: Optional[DataSource]
    targets: Optional[list[DataTarget]]
    infer_options: Optional[int]
    credentials: Credentials = Credentials()


class FeatureSetIngestOutput(FeatureStoreBaseModel):
    feature_set: FeatureSet
    run_object: dict



# Copyright 2018 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib
import inspect
from typing import Union

import mlrun.errors
from mlrun.config import config as mlconf
from mlrun.execution import MLClientCtx
from mlrun.track.tracker import Tracker
from mlrun.utils import logger
from mlrun.utils.singleton import Singleton

# Add a tracker to this list for it to be added into the global tracker manager (only if available in the interpreter):
_TRACKERS = ["mlflow"]

# A list for the available trackers during runtime. It will be setup at the beginning of the run by the function
# `_collect_available_trackers`:
_AVAILABLE_TRACKERS: list[Tracker] = None


class TrackerManager(metaclass=Singleton):
    """
    A class for handling multiple `Tracker` instances during a run. The manager is a singleton class and should be
    retrieved using the `mlrun.track.get_trackers_manager` function.
    """

    def __init__(self):
        """
        Initialize a new empty tracker manager.
        """
        self._trackers: list[Tracker] = []

        # Check general config for tracking usage, if false we return an empty manager
        if mlconf.external_platform_tracking.enabled:
            # Check if the available trackers were collected:
            if _AVAILABLE_TRACKERS is None:
                self._collect_available_trackers()

            # Add relevant trackers (enabled ones by the configuration) to be managed:
            for available_tracker in _AVAILABLE_TRACKERS:
                if available_tracker.is_enabled():
                    self.add_tracker(tracker=available_tracker())

    def add_tracker(self, tracker: Tracker):
        """
        Adds a Tracker to trackers list.

        :param tracker: The tracker class to add
        """
        # Check if tracker of same type already in manager
        if any(isinstance(t, type(tracker)) for t in self._trackers):
            logger.warn("Tracker already in manager", tracker=tracker)

        self._trackers.append(tracker)
        logger.debug("Added tracker", tracker=tracker.__class__.__name__)

    def pre_run(self, context: MLClientCtx):
        """
        Goes over all the managed trackers and calls their `pre_run` method.

        :param context: Active mlrun context.
        """
        for tracker in self._trackers:
            tracker.pre_run(context=context)

    def post_run(self, context: Union[MLClientCtx, dict]) -> Union[MLClientCtx, dict]:
        """
        Goes over all the managed trackers and calls their `post_run` method.

        :param context: Active mlrun context.

        :return: The context updated with the trackers products.
        """
        if not self._trackers:
            return context

        # Check if the context received is a dict to initialize it as an `MLClientCtx` object:
        is_context_dict = isinstance(context, dict)
        if is_context_dict:
            context = MLClientCtx.from_dict(
                context, include_status=True, store_run=False
            )

        # Go over the trackers and call the `post_run` method:
        for tracker in self._trackers:
            try:
                tracker.post_run(context)
            except Exception as e:
                logger.warn(
                    f"Tracker {tracker.__class__.__name__} failed in post run with the following exception",
                    exception=mlrun.errors.err_to_str(e),
                )

        # Commit changes:
        context.commit()

        # Return the context (cast to dict if received as a dict):
        return context.to_dict() if is_context_dict else context

    def _collect_available_trackers(self):
        """
        Set up the `_AVAILABLE_TRACKERS` list with trackers that were able to be imported.
        The tracked modules are not in MLRun's requirements and so it trys to import the module file of each and
        only if it succeeds (not raising `ModuleNotFoundError`) it collects it as an available tracker.
        """
        global _AVAILABLE_TRACKERS

        # Initialize an empty list:
        _AVAILABLE_TRACKERS = []

        # Go over the trackers and try to collect them:
        for tracker_module_name in _TRACKERS:
            # Try to import:
            try:
                tracker_module = importlib.import_module(
                    f"mlrun.track.trackers.{tracker_module_name}_tracker"
                )
            except ModuleNotFoundError:
                continue
            # Look for `Tracker` classes inside:
            _AVAILABLE_TRACKERS += [
                member
                for _, member in inspect.getmembers(
                    tracker_module,
                    lambda m: (
                        # Validate it is declared in the module:
                        hasattr(m, "__module__")
                        and m.__module__ == tracker_module.__name__
                        # Validate it is a `Tracker`:
                        and isinstance(m, type)
                        and issubclass(m, Tracker)
                        # Validate it is not a "protected" `Tracker`:
                        and not m.__name__.startswith("_")
                    ),
                )
            ]



# Copyright 2023 Iguazio
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import multiprocessing
import typing

from mlrun.utils import logger


def evaluate_condition_in_separate_process(
    condition: str, context: dict[str, typing.Any], timeout: int = 5
):
    if not condition:
        return True

    receiver, sender = multiprocessing.Pipe()
    p = multiprocessing.Process(
        target=_evaluate_condition_wrapper,
        args=(sender, condition, context),
    )
    p.start()
    if receiver.poll(timeout):
        result = receiver.recv()
        p.join()
        return result
    else:
        p.kill()
        logger.warning(
            f"Condition evaluation timed out after {timeout} seconds. Ignoring condition",
            condition=condition,
        )
        return True


def _evaluate_condition_wrapper(
    connection, condition: str, context: dict[str, typing.Any]
):
    connection.send(_evaluate_condition(condition, context))
    return connection.close()


def _evaluate_condition(condition: str, context: dict[str, typing.Any]):
    import jinja2.sandbox

    jinja_env = jinja2.sandbox.SandboxedEnvironment()
    template = jinja_env.from_string(condition)
    result = template.render(**context)
    if result.lower() in ["0", "no", "n", "f", "false", "off"]:
        return False

    # if the condition is not a boolean, we ignore the condition
    return True
